"""
Main class.
"""

import asyncio
import copy
import faulthandler
import json
import os
import pdb
import re
import sys
import time
import traceback
from dataclasses import MISSING
from operator import itemgetter
from pathlib import Path
from textwrap import indent

import litellm
import pyfiglet
import tldextract
from beartype.door import is_bearable
from beartype.typing import Any, Callable, List, Literal, Optional, Union
from langchain.docstore.document import Document
from langchain.globals import set_debug, set_llm_cache, set_verbose
from langchain.retrievers import ContextualCompressionRetriever
from langchain.retrievers.document_compressors import DocumentCompressorPipeline
from langchain.retrievers.merger_retriever import MergerRetriever
from langchain_community.document_transformers import EmbeddingsRedundantFilter
from langchain_community.llms import FakeListLLM
from langchain_community.retrievers import KNNRetriever, SVMRetriever
from langchain_core.output_parsers.string import StrOutputParser
from langchain_core.runnables import RunnablePassthrough, chain
from langchain_core.runnables.base import RunnableEach
from tqdm import tqdm

from .utils.batch_file_loader import batch_load_doc
from .utils.customs.fix_llm_caching import SQLiteCacheFixed
from .utils.embeddings import create_embeddings, load_embeddings_engine
from .utils.env import (
    WDOC_ALLOW_NO_PRICE,
    WDOC_DEBUGGER,
    WDOC_DEFAULT_EMBED_MODEL,
    WDOC_DEFAULT_MODEL,
    WDOC_DEFAULT_QUERY_EVAL_MODEL,
    WDOC_EMBED_TESTING,
    WDOC_LLM_MAX_CONCURRENCY,
    WDOC_OPEN_ANKI,
    WDOC_TYPECHECKING,
)
from .utils.errors import (
    NoDocumentsAfterLLMEvalFiltering,
    NoDocumentsRetrieved,
    ShouldIncreaseTopKAfterLLMEvalFiltering,
)

# just in case: setting the is_private flag as soon as possible to the env variables
from .utils.flags import is_debug, is_private, is_verbose, is_piped
from .utils.interact import ask_user
from .utils.llm import TESTING_LLM, load_llm

# import this first because it sets the logging level
from .utils.logger import (
    cache_dir,
    log_dir,
    logger,
    md_printer,
    red,
    set_help_md_as_docstring,
    set_parse_file_help_md_as_docstring,
    whi,
    yel,
)
from .utils.misc import (  # debug_chain,
    DocDict,
    ModelName,
    ankiconnect,
    average_word_length,
    check_docs_tkn_length,
    create_langfuse_callback,
    disable_internet,
    extra_args_types,
    get_model_price,
    get_splitter,
    get_supported_model_params,
    get_tkn_length,
    model_name_matcher,
    query_eval_cache,
    set_func_signature,
    thinking_answer_parser,
    wpm,
)
from .utils.prompts import prompts
from .utils.retrievers import create_multiquery_retriever, create_parent_retriever
from .utils.tasks.query import (
    check_intermediate_answer,
    collate_intermediate_answers,
    parse_eval_output,
    pbar_chain,
    pbar_closer,
    refilter_docs,
    semantic_batching,
    sieve_documents,
)
from .utils.tasks.summarize import do_summarize
from .utils.typechecker import optional_typecheck

logger.info("Starting wdoc")

os.environ["TOKENIZERS_PARALLELISM"] = "true"

# sensible default user agent
if "USER_AGENT" not in os.environ:
    os.environ["USER_AGENT"] = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/115.0"
    )


@optional_typecheck
@set_help_md_as_docstring
class wdoc:
    """
    This docstring is dynamically updated with the content of wdoc/docs/help.md
    """

    VERSION: str = "2.6.9"
    allowed_extra_args = extra_args_types
    md_printer = md_printer

    @optional_typecheck
    @set_func_signature
    def __init__(
        self,
        task: Literal["query", "search", "summarize", "summarize_then_query"],
        filetype: str = "auto",
        model: str = WDOC_DEFAULT_MODEL,
        embed_model: str = WDOC_DEFAULT_EMBED_MODEL,
        query_eval_model: Optional[str] = WDOC_DEFAULT_QUERY_EVAL_MODEL,
        embed_kwargs: Optional[dict] = None,
        save_embeds_as: Union[str, Path] = "{user_cache}/latest_docs_and_embeddings",
        load_embeds_from: Optional[Union[str, Path]] = None,
        top_k: Union[str, int] = "auto_200_500",
        query: Optional[str] = None,
        query_retrievers: str = "default_multiquery",
        query_eval_check_number: int = 3,
        query_relevancy: Union[float, int] = 0.0,
        summary_n_recursion: int = 0,
        summary_language: str = "the same language as the document",  # <- the LLM will understand
        llm_verbosity: Union[bool, int] = False,
        debug: Union[bool, int] = False,
        verbose: Union[bool, int] = False,
        dollar_limit: int = 5,
        notification_callback: Optional[Callable] = None,
        disable_llm_cache: Union[bool, int] = False,
        file_loader_parallel_backend: Literal[
            "loky", "threading", "multiprocessing"
        ] = "loky",
        file_loader_n_jobs: int = -1,
        private: Union[bool, int] = False,
        llms_api_bases: Optional[Union[dict, str]] = None,
        import_mode: Union[bool, int] = False,
        disable_md_printing: bool = is_piped,
        oneoff: bool = False,
        silent: bool = False,
        version: bool = False,
        **cli_kwargs,
    ) -> None:
        """
        This docstring is dynamically updated with the content of wdoc/docs/help.md
        """
        if version:
            print(self.VERSION)
            return
        if notification_callback is not None:

            @optional_typecheck
            def ntfy(text: str) -> str:
                out = notification_callback(text)
                assert (
                    out == text
                ), "The notification callback must return the same string"
                return out

            ntfy("Starting wdoc")
        else:

            @optional_typecheck
            def ntfy(text: str) -> str:
                return text

        self.ntfy = ntfy

        if debug or WDOC_DEBUGGER:
            debug_exceptions(instance=self)

        elif notification_callback:

            def print_exception(exc_type, exc_value, exc_traceback):
                if not issubclass(exc_type, KeyboardInterrupt):
                    message = "An error has occured:\n"
                    message += "\n".join(
                        [line for line in traceback.format_tb(exc_traceback)]
                    )
                    message += "\n" + str(exc_type) + " : " + str(exc_value)
                    self.ntfy(message)
                    sys.exit(1)

            sys.excepthook = print_exception
            faulthandler.enable()

        red(pyfiglet.figlet_format("wdoc"))

        # make sure the extra args are valid
        for k in cli_kwargs:
            if k not in self.allowed_extra_args:
                raise Exception(
                    red(
                        f"Found unexpected keyword argument: '{k}'\nThe allowed arguments are {','.join(self.allowed_extra_args)}"
                    )
                )

            # type checking of extra args
            if WDOC_TYPECHECKING in ["crash", "warn"]:
                val = cli_kwargs[k]
                # curr_type = type(val)
                expected_type = self.allowed_extra_args[k]
                if expected_type is str:
                    assert val.strip(), f"Empty string found for cli_kwargs: '{k}'"
                if isinstance(val, list):
                    assert val, f"Empty list found for cli_kwargs: '{k}'"
                if not is_bearable(val, expected_type):
                    raise Exception(
                        f"Cli_kwargs '{k}' is of type '{type(val)}' instead of '{expected_type}'"
                    )

        if model == TESTING_LLM:
            if model != TESTING_LLM:
                red(
                    f"Detected 'testing' model in {model}, setting it to '{TESTING_LLM}'"
                )
                model = TESTING_LLM
            else:
                red(f"Detected 'testing' model in {model}")
            if isinstance(query_eval_model, str):
                if query_eval_model != TESTING_LLM:
                    red(f"Setting the query_eval_model to {TESTING_LLM} too")
                    query_eval_model = TESTING_LLM

        # checking argument validity
        assert (
            "loaded_docs" not in cli_kwargs
        ), "'loaded_docs' cannot be an argument as it is used internally"
        assert (
            "loaded_embeddings" not in cli_kwargs
        ), "'loaded_embeddings' cannot be an argument as it is used internally"
        task = task.replace("summary", "summarize")
        assert task in [
            "query",
            "search",
            "summarize",
            "summarize_then_query",
        ], "invalid task value"
        if task in ["summarize", "summarize_then_query"]:
            assert not load_embeds_from, "can't use load_embeds_from if task is summary"
        if task in ["query", "search", "summarize_then_query"]:
            assert (
                query_eval_model is not None
            ), "query_eval_model can't be None if doing RAG"
        else:
            query_eval_model = None
        if filetype == "auto":
            assert (
                "path" in cli_kwargs and cli_kwargs["path"]
            ), "If filetype is 'auto', a --path must be given"
        elif filetype == "string":
            cli_kwargs["path"] = "empty placeholder"
        assert (
            "/" in model
        ), "model must be in litellm format: provider/model. For example 'openai/gpt-4o'"
        if model != TESTING_LLM and model.split("/", 1)[0] not in list(
            litellm.models_by_provider.keys()
        ):
            raise Exception(
                f"For model '{model}': backend not found in "
                "litellm nor 'testing'.\nList of litellm providers/backend:\n"
                f"{litellm.models_by_provider.keys()}"
            )
        if embed_kwargs is None:
            embed_kwargs = {}
        if isinstance(embed_kwargs, str):
            try:
                embed_kwargs = json.loads(embed_kwargs)
            except Exception as err:
                raise Exception(
                    f"Failed to parse embed_kwargs: '{embed_kwargs}'"
                ) from err
        assert isinstance(embed_kwargs, dict), f"Not a dict but {type(embed_kwargs)}"
        assert query_eval_check_number > 0, "query_eval_check_number value"

        if llms_api_bases is None:
            llms_api_bases = {}
        elif isinstance(llms_api_bases, str):
            try:
                llms_api_bases = json.loads(llms_api_bases)
            except Exception as err:
                raise Exception(f"Error when parsing llms_api_bases as a dict: {err}")
        assert isinstance(
            llms_api_bases, dict
        ), "llms_api_bases must be a dict or be a string that can be parsed as a dict"
        for k in llms_api_bases:
            assert k in [
                "model",
                "query_eval_model",
                "embeddings",
            ], f"Invalid k of llms_api_bases not in 'model', 'query_eval_model', 'embeddings': {k}"
        for k in ["model", "query_eval_model", "embeddings"]:
            if k not in llms_api_bases:
                llms_api_bases[k] = None
        if (
            llms_api_bases["model"] == llms_api_bases["query_eval_model"]
            and llms_api_bases["model"] == llms_api_bases["embeddings"]
            and llms_api_bases["model"]
        ):
            red(
                "Setting litellm wide api_base because it's the same for model, query_eval_model and embeddings"
            )
            litellm.api_base = llms_api_bases["model"]
        assert isinstance(
            private, bool
        ), "private arg should be a boolean, not {private}"
        assert private == is_private
        if private:
            assert llms_api_bases[
                "model"
            ], "private is set but llms_api_bases['model'] is not set"
            assert llms_api_bases[
                "query_eval_model"
            ], "private is set but llms_api_bases['query_eval_model'] is not set"
            os.environ["WDOC_PRIVATE_MODE"] = "true"
            for k in dict(os.environ):
                if k.endswith("_API_KEY") or k.endswith("_API_KEYS"):
                    red(
                        f"private mode enabled: overwriting '{k}' from environment variables just in case"
                    )
                    os.environ[k] = "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"

            if any("LANGFUSE_" in k for k in os.environ.keys()):
                red("Disabled langfuse because using private mode.")
            os.environ["LANGFUSE_PUBLIC_KEY"] = "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"
            os.environ["LANGFUSE_SECRET_KEY"] = "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"
            os.environ["LANGFUSE_HOST"] = "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"
            os.environ["WDOC_LANGFUSE_PUBLIC_KEY"] = (
                "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"
            )
            os.environ["WDOC_LANGFUSE_SECRET_KEY"] = (
                "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"
            )
            os.environ["WDOC_LANGFUSE_HOST"] = "REDACTED_BECAUSE_WDOC_IN_PRIVATE_MODE"
            if litellm.success_callback or litellm.failure_callback:
                red("Disabled litellm callbacks because using private mode.")
            litellm.success_callback = []
            litellm.failure_callback = []

            # to be extra safe, let's try to block any remote connection
            disable_internet(
                allowed=llms_api_bases,
            )

        else:
            os.environ["WDOC_PRIVATE_MODE"] = "false"

        if (model != TESTING_LLM) and (not llms_api_bases["model"]):
            model = model_name_matcher(model)
        if (query_eval_model is not None) and (not llms_api_bases["query_eval_model"]):
            if model == TESTING_LLM:
                assert query_eval_model == TESTING_LLM
            else:
                query_eval_model = model_name_matcher(query_eval_model)

        if query is True:
            # otherwise specifying --query and forgetting to add text fails
            query = None

        if isinstance(query, str):
            query = query.strip() or None
        assert file_loader_parallel_backend in [
            "loky",
            "threading",
            "multiprocessing",
        ], "Invalid value for file_loader_parallel_backend"
        assert isinstance(file_loader_n_jobs, int), "file_loader_n_jobs mus be an int"
        if "{user_cache}" in save_embeds_as:
            save_embeds_as = save_embeds_as.replace("{user_cache}", str(cache_dir))
        if query_relevancy is None:
            query_relevancy = 0.0
        query_relevancy = float(query_relevancy)

        # parsing top_k value
        if isinstance(top_k, str):
            try:
                starting_top_k, max_top_k = top_k.split("_")[1:]
                starting_top_k = int(starting_top_k)
                max_top_k = int(max_top_k)
                assert max_top_k > starting_top_k, "M<=N"
                assert starting_top_k > 0, "N<=0"

            except Exception as err:
                raise Exception(
                    "Failed to parse string top_k value. If top_k "
                    "is a string, the expected format is 'auto_N_M' with N "
                    f"and M ints and M>N. Received: {top_k}"
                ) from err
            top_k = starting_top_k
            self.max_top_k = max_top_k
        else:
            self.max_top_k = None

        # storing as attributes
        self.model = ModelName(model)
        if query_eval_model is not None:
            self.query_eval_model = ModelName(query_eval_model)
        else:
            self.query_eval_model = None
        self.task = task
        self.filetype = filetype
        self.embed_model = ModelName(embed_model)
        self.embed_kwargs = embed_kwargs
        self.save_embeds_as = save_embeds_as
        self.load_embeds_from = load_embeds_from
        self.top_k = top_k
        self.query_retrievers = (
            query_retrievers
            if model != TESTING_LLM
            else query_retrievers.replace("multiquery", "")
        )
        self.query_eval_check_number = int(query_eval_check_number)
        self.query_relevancy = query_relevancy
        self.debug = debug
        self.verbose = verbose
        self.cli_kwargs = cli_kwargs
        self.llm_verbosity = llm_verbosity
        self.summary_n_recursion = summary_n_recursion
        self.summary_language = summary_language
        self.dollar_limit = dollar_limit
        self.private = bool(private)
        self.disable_llm_cache = bool(disable_llm_cache)
        self.file_loader_parallel_backend = file_loader_parallel_backend
        self.file_loader_n_jobs = file_loader_n_jobs
        self.llms_api_bases = llms_api_bases
        self.import_mode = import_mode
        self.oneoff = oneoff

        if disable_llm_cache:
            self.llm_cache = False
        else:
            if not private:
                self.llm_cache = SQLiteCacheFixed(
                    database_path=(cache_dir / "langchain_db").resolve().absolute(),
                    verbose=is_verbose,
                )
            else:
                self.llm_cache = SQLiteCacheFixed(
                    database_path=(cache_dir / "private_langchain_db")
                    .resolve()
                    .absolute(),
                    verbose=is_verbose,
                )
            set_llm_cache(self.llm_cache)

        if WDOC_ALLOW_NO_PRICE:
            red(
                f"Disabling price computation for {self.model.original} because env var 'WDOC_ALLOW_NO_PRICE' is 'true'"
            )
            self.llm_price = [0.0, 0.0]

        elif llms_api_bases["model"]:
            red(
                f"Disabling price computation for model because api_base for 'model' was modified to {llms_api_bases['model']}"
            )
            self.llm_price = [0.0, 0.0]
        elif self.model.backend == TESTING_LLM:
            red(
                f"Disabling price computation for model because api_base for 'model' was modified to {llms_api_bases['model']}"
            )
            self.llm_price = [0.0, 0.0]
        else:
            self.llm_price = get_model_price(self.model.original)

        if self.query_eval_model is not None:
            if WDOC_ALLOW_NO_PRICE:
                red(
                    f"Disabling price computation for {self.query_eval_model.original} because env var 'WDOC_ALLOW_NO_PRICE' is 'true'"
                )
                self.query_evalllm_price = [0.0, 0.0]
            elif llms_api_bases["query_eval_model"]:
                red(
                    "Disabling price computation for query_eval_model because api_base was modified"
                )
                self.query_evalllm_price = [0.0, 0.0]
            else:
                self.query_evalllm_price = get_model_price(
                    self.query_eval_model.original
                )

        if is_verbose:
            set_verbose(True)
            litellm.set_verbose = True

            llm_verbosity = True
            whi(f"Cache location: {cache_dir.absolute()}")
            whi(f"Log location: {log_dir.absolute()}")
        else:
            set_verbose(False)
            set_debug(False)
            litellm.set_verbose = False
            # fix from https://github.com/BerriAI/litellm/issues/2256
            import logging

            for logger_name in [
                "LiteLLM Proxy",
                "LiteLLM Router",
                "LiteLLM",
                "litellm",
                "httpx",
            ]:
                logger = logging.getLogger(logger_name)
                # logger.setLevel(logging.CRITICAL + 1)
                logger.setLevel(logging.WARNING)
            for logger_name in ["bs4"]:
                logger = logging.getLogger(logger_name)
                logger.setLevel(logging.CRITICAL)
        if debug:
            assert is_verbose
            assert is_debug
            # os.environ["LANGCHAIN_TRACING_V2"] = "true"
            set_debug(True)

        # don't crash if extra arguments are used for a model
        # litellm.drop_params = True  # drops parameters that are not used by some models

        # compile include / exclude regex
        if "include" in self.cli_kwargs:
            for i, inc in enumerate(self.cli_kwargs["include"]):
                if inc == inc.lower():
                    self.cli_kwargs["include"][i] = re.compile(inc, flags=re.IGNORECASE)
                else:
                    self.cli_kwargs["include"][i] = re.compile(inc)
        if "exclude" in self.cli_kwargs:
            for i, exc in enumerate(self.cli_kwargs["exclude"]):
                if exc == exc.lower():
                    self.cli_kwargs["exclude"][i] = re.compile(exc, flags=re.IGNORECASE)
                else:
                    self.cli_kwargs["exclude"][i] = re.compile(exc)

        # loading llm
        self.llm = load_llm(
            modelname=self.model,
            backend=self.model.backend,
            llm_cache=self.llm_cache,
            temperature=0,
            llm_verbosity=self.llm_verbosity,
            api_base=self.llms_api_bases["model"],
            private=self.private,
            tags=["strong_model"],
        )
        # if "anthropic" in self.model.lower() or "anthropic" in self.backend.lower():
        #     prompts.enable_prompt_caching("answer")
        #     prompts.enable_prompt_caching("combine")
        #     prompts.enable_prompt_caching("multiquery")

        # loading documents
        if not load_embeds_from:
            # remove args that are indented only for the instanciation and
            # not as loading argument
            filtered_cli_kwargs = self.cli_kwargs.copy()
            for k in [
                "embed_instruct",
                "filter_content",
                "filter_metadata",
                "out_file",
            ]:
                if k in filtered_cli_kwargs:
                    del filtered_cli_kwargs[k]

            self.loaded_docs = batch_load_doc(
                llm_name=self.model,
                filetype=self.filetype,
                task=self.task,
                backend=self.file_loader_parallel_backend,
                n_jobs=self.file_loader_n_jobs if not is_debug else 1,
                **filtered_cli_kwargs,
            )
        else:
            self.loaded_docs = None  # will be loaded when embeddings are loaded

        if self.task in ["query", "search", "summary_then_query"]:
            self.prepare_query_task()

        if self.import_mode:
            if is_verbose:
                whi(
                    "Ready to query or summarize, call your_instance.query_task(your_question)"
                )
            return

        if self.task in ["summarize", "summarize_then_query"]:
            self.summary_task()

            if self.task == "summary_then_query":
                whi("Done summarizing. Switching to query mode.")
                if "logit_bias" in get_supported_model_params(self.model):
                    del self.llm.model_kwargs["logit_bias"]
            else:
                whi("Done summarizing.")
                return

        else:
            assert self.task in [
                "query",
                "search",
                "summary_then_query",
            ], f"Invalid task: {self.task}"
            while True:
                self.query_task(query=query)
                query = None

    @optional_typecheck
    def summary_task(self) -> dict:
        docs_tkn_cost = {}
        for doc in self.loaded_docs:
            meta = doc.metadata["path"]
            if meta not in docs_tkn_cost:
                docs_tkn_cost[meta] = get_tkn_length(doc.page_content)
            else:
                docs_tkn_cost[meta] += get_tkn_length(doc.page_content)

        full_tkn = sum(list(docs_tkn_cost.values()))
        red("Token price of each document:")
        for k, v in docs_tkn_cost.items():
            pr = v * (self.llm_price[0] * 4 + self.llm_price[1]) / 5
            red(f"- {v:>6}: {k:>10} - ${pr:04f}")

        red(f"Total number of tokens in documents to summarize: '{full_tkn}'")
        # use an heuristic to estimate the price to summarize
        adj_price = (self.llm_price[0] * 3 + self.llm_price[1] * 2) / 5
        estimate_dol = full_tkn * adj_price / 100 * 1.2
        if self.summary_n_recursion:
            for i in range(1, self.summary_n_recursion + 1):
                estimate_dol += full_tkn * ((2 / 5) ** i) * adj_price / 100 * 1.2
        whi(
            self.ntfy(
                f"Estimate of the LLM cost to summarize: ${estimate_dol:.4f} for {full_tkn} tokens."
            )
        )
        if estimate_dol > self.dollar_limit:
            if self.llms_api_bases["model"]:
                raise Exception(
                    red(
                        self.ntfy(
                            f"Cost estimate ${estimate_dol:.5f} > ${self.dollar_limit} which is absurdly high. Has something gone wrong? Quitting."
                        )
                    )
                )
            else:
                red(
                    "Cost estimate > limit but the api_base was modified so not crashing."
                )

        llm_params = get_supported_model_params(self.model)
        if "logit_bias" in llm_params:
            # increase likelyhood that chatgpt will use indentation by
            # biasing towards adding space.
            logit_val = 3
            self.llm.model_kwargs["logit_bias"] = {
                12: logit_val,  # '-'
                # 220: logit_val,  # ' '
                # 532: logit_val,  # ' -'
                # 9: logit_val,  # '*'
                # 1635: logit_val,  # ' *'
                # 197: logit_val,  # '\t'
                334: logit_val,  # '**'
                # 25: logit_val,  # ':'
                # 551: logit_val,  # ' :'
                # 13: -1,  # '.'
                # logit bias for indentation, the number of space, because it consumes less token than using \t
                257: logit_val,  # "    "
                260: logit_val,  # "        "
                1835: logit_val,  # "            "
                338: logit_val,  # "                "
                3909: logit_val,  # "                    "
                5218: logit_val,  # "                        "
                6663: logit_val,  # "                            "
                792: logit_val,  # "                                "
                10812: logit_val,  # "                                    "
                13137: logit_val,  # "                                        "
                15791: logit_val,  # "                                            "
                19273: logit_val,  # "                                                "
                25343: logit_val,  # "                                                    "
                29902: logit_val,  # "                                                        "
                39584: logit_val,  # "                                                            "
                5341: logit_val,  # "                                                                "
                52168: logit_val,  # "                                                                    "
                38244: logit_val,  # "                                                                        "
                56899: logit_val,  # "                                                                            "
                98517: logit_val,  # "                                                                                "
            }
        if "frequency_penalty" in llm_params:
            self.llm.model_kwargs["frequency_penalty"] = 0.0
        if "presence_penalty" in llm_params:
            self.llm.model_kwargs["presence_penalty"] = 0.0
        if "temperature" in llm_params:
            self.llm.model_kwargs["temperature"] = 0.0

        @optional_typecheck
        def summarize_documents(
            path: Any,
            relevant_docs: List,
        ) -> dict:
            assert relevant_docs, "Empty relevant_docs!"

            # parse metadata from the doc
            metadata = []
            if "http" in path:
                item_name = tldextract.extract(path).registered_domain
            elif "/" in path and Path(path).exists():
                item_name = Path(path).name
            else:
                item_name = path

            if "title" in relevant_docs[0].metadata:
                item_name = (
                    f"{relevant_docs[0].metadata['title'].strip()} - {item_name}"
                )
            else:
                metadata.append(f"<title>\n{item_name.strip()}\n</title>")

            # replace # in title as it would be parsed as a tag
            item_name = item_name.replace("#", r"\#")

            if "doc_reading_time" in relevant_docs[0].metadata:
                doc_reading_length = relevant_docs[0].metadata["doc_reading_time"]
                metadata.append(
                    f"<reading_length>\n{doc_reading_length:.1f} minutes\n</reading_length>"
                )
            else:
                doc_reading_length = 0
            if "author" in relevant_docs[0].metadata:
                author = relevant_docs[0].metadata["author"].strip()
                metadata.append(f"<author>\n{author}\n</author>")
            else:
                author = None
            if "yt_chapters" in relevant_docs[0].metadata:
                chapters = json.dumps(relevant_docs[0].metadata["yt_chapters"])
                metadata.append(f"<youtube_chapters>\n{chapters}\n</youtube_chapters>")

            if metadata:
                metadata = "<text_metadata>\n" + "\n".join(metadata) + "\n"
                metadata += "<section_number>\n[PROGRESS]\n</section_number>\n"
                metadata += "</text_metadata>"
            else:
                metadata = "<text_metadata><section_number>[PROGRESS]</section_number></text_metadata>"

            # summarize each chunk of the link and return one text
            (
                summary,
                n_chunk,
                doc_total_cost,
                doc_total_tokens_in,
                doc_total_tokens_out,
            ) = do_summarize(
                docs=relevant_docs,
                metadata=metadata,
                language=self.summary_language,
                modelbackend=self.model.backend,
                llm=self.llm,
                llm_price=self.llm_price,
                verbose=self.llm_verbosity,
            )

            # get reading length of the summary
            real_text = "".join(
                [letter for letter in list(summary) if letter.isalpha()]
            )
            sum_reading_length = len(real_text) / average_word_length / wpm
            whi(f"{item_name} reading length is {sum_reading_length:.1f}")

            recursive_summaries = {0: summary}
            prev_real_text = MISSING
            if self.summary_n_recursion > 0:
                for n_recur in range(1, self.summary_n_recursion + 1):
                    summary_text = copy.deepcopy(recursive_summaries[n_recur - 1])
                    if not self.import_mode:
                        red(f"Doing summary check #{n_recur} of {item_name}")

                    # remove any chunk count that is not needed to summarize
                    sp = summary_text.split("\n")
                    for i, l in enumerate(sp):
                        if l.strip() == "- ---":
                            sp[i] = None
                        elif re.search(r"- Chunk \d+/\d+", l):
                            sp[i] = None
                        elif l.strip().startswith("- BEFORE RECURSION #"):
                            for new_i in range(i, len(sp)):
                                sp[new_i] = None
                            break
                    summary_text = "\n".join([s.rstrip() for s in sp if s])
                    assert "- ---" not in summary_text, "Found chunk separator"
                    assert "- Chunk " not in summary_text, "Found chunk marker"
                    assert (
                        "- BEFORE RECURSION # " not in summary_text
                    ), "Found recursion block"

                    splitter = get_splitter(
                        "recursive_summary",
                        modelname=self.model,
                    )
                    summary_docs = [Document(page_content=summary_text)]
                    summary_docs = splitter.transform_documents(summary_docs)
                    assert summary_docs != relevant_docs
                    try:
                        check_docs_tkn_length(summary_docs, item_name)
                    except Exception as err:
                        red(
                            f"Exception when checking if {item_name} could be recursively summarized for the #{n_recur} time: {err}"
                        )
                        break
                    (
                        summary_text,
                        n_chunk,
                        new_doc_total_cost,
                        new_doc_total_tokens_in,
                        new_doc_total_tokens_out,
                    ) = do_summarize(
                        docs=summary_docs,
                        metadata=metadata,
                        language=self.summary_language,
                        modelbackend=self.model.backend,
                        llm=self.llm,
                        llm_price=self.llm_price,
                        verbose=self.llm_verbosity,
                        n_recursion=n_recur,
                    )
                    doc_total_tokens_in += new_doc_total_tokens_in
                    doc_total_tokens_out += new_doc_total_tokens_out
                    doc_total_cost += new_doc_total_cost

                    # clean text again to compute the reading length
                    sp = summary_text.split("\n")
                    for i, l in enumerate(sp):
                        if l.strip() == "- ---":
                            sp[i] = None
                        elif re.search(r"- Chunk \d+/\d+", l):
                            sp[i] = None
                        elif l.strip().startswith("- BEFORE RECURSION #"):
                            for new_i in range(i, len(sp)):
                                sp[new_i] = None
                            break
                    real_text = "\n".join([s.rstrip() for s in sp if s])
                    assert "- ---" not in real_text, "Found chunk separator"
                    assert "- Chunk " not in real_text, "Found chunk marker"
                    assert (
                        "- BEFORE RECURSION # " not in real_text
                    ), "Found recursion block"
                    real_text = "".join(
                        [letter for letter in list(real_text) if letter.isalpha()]
                    )
                    sum_reading_length = len(real_text) / average_word_length / wpm
                    whi(
                        f"{item_name} reading length after recursion #{n_recur} is {sum_reading_length:.1f}"
                    )
                    if prev_real_text is not MISSING:
                        if real_text == prev_real_text:
                            if not self.import_mode:
                                red(
                                    f"Identical summary after {n_recur} "
                                    "recursion, adding more recursion will not "
                                    "help so stopping here"
                                )
                            recursive_summaries[n_recur] = summary_text
                            break
                    prev_real_text = real_text

                    assert n_recur not in recursive_summaries
                    if summary_text not in recursive_summaries:
                        if not self.import_mode:
                            red(
                                f"Identical summary after {n_recur} "
                                "recursion, adding more recursion will not "
                                "help so stopping here"
                            )
                        recursive_summaries[n_recur] = summary_text
                        break
                    else:
                        recursive_summaries[n_recur] = summary_text

            best_sum_i = max(list(recursive_summaries.keys()))
            if not self.import_mode:
                print("\n\n")
                md_printer("# Summary")
                md_printer(f"## {path}")
                md_printer(recursive_summaries[best_sum_i])

                doc_total_tokens = doc_total_tokens_in + doc_total_tokens_out
                red(
                    f"Tokens used for {path}: '{doc_total_tokens}' (in: {doc_total_tokens_in}, out: {doc_total_tokens_out}, cost: ${doc_total_cost:.5f})"
                )

            summary_tkn_length = get_tkn_length(recursive_summaries[best_sum_i])

            header = (
                f"\n- {item_name}    cost: {doc_total_tokens} (${doc_total_cost:.5f})"
            )
            if doc_reading_length:
                header += f"    {doc_reading_length:.1f} minutes"
            if author:
                header += f"    by '{author}'"
            header += f"    original path: '{path}'"
            header += f"    wdoc version {self.VERSION} with model {self.model}"

            # save to output file
            if "out_file" in self.cli_kwargs:
                assert not self.import_mode, "Can't use import_mode with --out_file"
                for nrecur, sum in recursive_summaries.items():
                    outfile = Path(self.cli_kwargs["out_file"])
                    if len(recursive_summaries) > 1 and nrecur < max(
                        list(recursive_summaries.keys())
                    ):
                        # also store intermediate summaries if present
                        outfile = outfile.parent / (outfile.stem + f".{nrecur + 1}.md")

                    with open(str(outfile), "a") as f:
                        if outfile.exists() and outfile.read_text().strip():
                            f.write("\n\n\n")
                        f.write(header)
                        if len(recursive_summaries) > 1:
                            f.write(
                                f"\n    Recursive summary pass: {nrecur + 1}/{len(recursive_summaries)}"
                            )

                        for bulletpoint in sum.split("\n"):
                            f.write("\n")
                            bulletpoint = bulletpoint.rstrip()
                            f.write(f"    {bulletpoint}")

            return {
                "path": path,
                "sum_reading_length": sum_reading_length,
                "sum_tkn_length": summary_tkn_length,
                "doc_reading_length": doc_reading_length,
                "doc_total_tokens": doc_total_tokens,
                "doc_total_cost": doc_total_cost,
                "summary": recursive_summaries[best_sum_i],
                "recursive_summaries": recursive_summaries,
                "author": author,
                "n_chunk": n_chunk,
            }

        results = summarize_documents(
            path=self.cli_kwargs["path"],
            relevant_docs=self.loaded_docs,
        )

        if not self.import_mode:
            red(
                self.ntfy(
                    f"Total cost of those summaries: {results['doc_total_tokens']} tokens for ${results['doc_total_cost']:.5f} (estimate was ${estimate_dol:.5f})"
                )
            )
            red(
                self.ntfy(
                    f"Total time saved by those summaries: {results['doc_reading_length']:.1f} minutes"
                )
            )
        else:
            self.ntfy(
                f"Total cost of those summaries: '{results['doc_total_tokens']}' (${results['doc_total_cost']:.5f}, estimate was ${estimate_dol:.5f})"
            )
            self.ntfy(
                f"Total time saved by those summaries: {results['doc_reading_length']:.1f} minutes"
            )

        llmcallback = self.llm.callbacks[0]
        total_cost = (
            self.llm_price[0] * llmcallback.prompt_tokens
            + self.llm_price[1] * llmcallback.completion_tokens
        )
        if llmcallback.total_tokens != results["doc_total_tokens"]:
            red(
                f"Cost discrepancy? Tokens used according to the callback: '{llmcallback.total_tokens}' (${total_cost:.5f})"
            )
        self.summary_results = results
        self.latest_cost = total_cost
        return results

    @optional_typecheck
    def prepare_query_task(self) -> None:
        # set argument that are better suited for querying
        if "logit_bias" in get_supported_model_params(self.model):
            # increase likelyhood that chatgpt will use indentation by
            # biasing towards adding space.
            logit_val = 3
            self.llm.model_kwargs["logit_bias"] = {
                12: logit_val,  # '-'
                # 220: logit_val,  # ' '
                # 532: logit_val,  # ' -'
                # 9: logit_val,  # '*'
                # 1635: logit_val,  # ' *'
                # 197: logit_val,  # '\t'
                334: logit_val,  # '**'
                # 25: logit_val,  # ':'
                # 551: logit_val,  # ' :'
                # 13: -1,  # '.'
                # logit bias for indentation, the number of space, because it consumes less token than using \t
                257: logit_val,  # "    "
                260: logit_val,  # "        "
                1835: logit_val,  # "            "
                338: logit_val,  # "                "
                3909: logit_val,  # "                    "
                5218: logit_val,  # "                        "
                6663: logit_val,  # "                            "
                792: logit_val,  # "                                "
                10812: logit_val,  # "                                    "
                13137: logit_val,  # "                                        "
                15791: logit_val,  # "                                            "
                19273: logit_val,  # "                                                "
                25343: logit_val,  # "                                                    "
                29902: logit_val,  # "                                                        "
                39584: logit_val,  # "                                                            "
                5341: logit_val,  # "                                                                "
                52168: logit_val,  # "                                                                    "
                38244: logit_val,  # "                                                                        "
                56899: logit_val,  # "                                                                            "
                98517: logit_val,  # "                                                                                "
            }
        if "frequency_penalty" in get_supported_model_params(self.model):
            self.llm.model_kwargs["frequency_penalty"] = 0.0
        if "presence_penalty" in get_supported_model_params(self.model):
            self.llm.model_kwargs["presence_penalty"] = 0.0
        if "temperature" in get_supported_model_params(self.model):
            self.llm.model_kwargs["temperature"] = 0.0

        # load embeddings for querying
        self.embedding_engine = load_embeddings_engine(
            modelname=self.embed_model,
            cli_kwargs=self.cli_kwargs,
            api_base=self.llms_api_bases["embeddings"],
            embed_kwargs=self.embed_kwargs,
            private=self.private,
            do_test=WDOC_EMBED_TESTING,
        )
        self.loaded_embeddings = create_embeddings(
            modelname=self.embed_model,
            cached_embeddings=self.embedding_engine,
            load_embeds_from=self.load_embeds_from,
            save_embeds_as=self.save_embeds_as,
            loaded_docs=self.loaded_docs,
            dollar_limit=self.dollar_limit,
            private=self.private,
        )

        # set default ask_user argument
        self.interaction_settings = {
            "top_k": self.top_k,
            "multiline": False,
            "retriever": self.query_retrievers,
            "task": self.task,
            "relevancy": self.query_relevancy,
        }
        self.all_texts = [
            v.page_content for k, v in self.loaded_embeddings.docstore._dict.items()
        ]

        # parse filters as callable for faiss filtering
        if "filter_metadata" in self.cli_kwargs or "filter_content" in self.cli_kwargs:
            if "filter_metadata" in self.cli_kwargs:
                # get the list of all metadata to see if a filter was not misspelled
                all_metadata_keys = set()
                for doc in tqdm(
                    self.loaded_embeddings.docstore._dict.values(),
                    desc="gathering metadata keys",
                    unit="doc",
                    disable=(not is_verbose) or is_piped,
                ):
                    for k in doc.metadata.keys():
                        all_metadata_keys.add(k)
                assert (
                    all_metadata_keys
                ), "No metadata keys found in any metadata, something went wrong!"

                if isinstance(self.cli_kwargs["filter_metadata"], str):
                    filter_metadata = self.cli_kwargs["filter_metadata"].split(",")
                else:
                    filter_metadata = self.cli_kwargs["filter_metadata"]
                assert isinstance(
                    filter_metadata, list
                ), f"filter_metadata must be a list, not {self.cli_kwargs['filter_metadata']}"

                # storing fast as list then in tupples for faster iteration
                filters_k_plus = []
                filters_k_minus = []
                filters_v_plus = []
                filters_v_minus = []
                filters_b_plus_keys = []
                filters_b_plus_values = []
                filters_b_minus_keys = []
                filters_b_minus_values = []
                for f in filter_metadata:
                    assert isinstance(f, str), f"Filter must be a string: '{f}'"
                    kvb = f[0]
                    assert kvb in [
                        "k",
                        "v",
                        "b",
                    ], f"filter 1st character must be k, v or b: '{f}'"
                    incexc = f[1]
                    assert incexc in [
                        "+",
                        "-",
                    ], f"filter 2nd character must be + or -: '{f}'"
                    incexc_str = "plus" if incexc == "+" else "minus"
                    assert f[2:].strip(), f"Filter can't be an empty regex: '{f}'"
                    pattern = f[2:].strip()
                    if kvb == "b":
                        assert ":" in f, (
                            "Filter starting with b must contain "
                            "a ':' to distinguish the key regex and the value "
                            f"regex: '{f}'"
                        )
                        key_pat, value_pat = pattern.split(":", 1)
                        if key_pat == key_pat.lower():
                            key_pat = re.compile(key_pat, flags=re.IGNORECASE)
                        else:
                            key_pat = re.compile(key_pat)
                        if value_pat == value_pat.lower():
                            value_pat = re.compile(value_pat, flags=re.IGNORECASE)
                        else:
                            value_pat = re.compile(value_pat)
                        assert (
                            key_pat not in locals()[f"filters_b_{incexc_str}_keys"]
                        ), (
                            f"Can't use several filters for the same key "
                            "regex. Use a single but more complex regex"
                            f": '{f}'"
                        )
                        locals()[f"filters_b_{incexc_str}_keys"].append(key_pat)
                        locals()[f"filters_b_{incexc_str}_values"].append(value_pat)
                    else:
                        if pattern == pattern.lower():
                            pattern = re.compile(pattern, flags=re.IGNORECASE)
                        else:
                            pattern = re.compile(pattern)
                        locals()[f"filters_{kvb}_{incexc_str}"].append(pattern)
                assert len(filters_b_plus_keys) == len(filters_b_plus_values)
                assert len(filters_b_minus_keys) == len(filters_b_minus_values)

                # store as tuple for faster iteration
                filters_k_plus = tuple(filters_k_plus)
                filters_k_minus = tuple(filters_k_minus)
                filters_v_plus = tuple(filters_v_plus)
                filters_v_minus = tuple(filters_v_minus)
                filters_b_plus_keys = tuple(filters_b_plus_keys)
                filters_b_plus_values = tuple(filters_b_plus_values)
                filters_b_minus_keys = tuple(filters_b_minus_keys)
                filters_b_minus_values = tuple(filters_b_minus_values)

                # check that all key filter indeed match metadata keys
                for k in (
                    filters_k_plus
                    + filters_k_minus
                    + filters_b_plus_keys
                    + filters_b_minus_keys
                ):
                    assert any(
                        k.match(str(key)) for key in all_metadata_keys
                    ), f"Key {k} didn't match any key in the metadata"

                @optional_typecheck
                def filter_meta(meta: dict) -> bool:
                    # match keys
                    for inc in filters_k_plus:
                        if not any(inc.match(str(k)) for k in meta.keys()):
                            return False
                    for exc in filters_k_minus:
                        if any(exc.match(str(k)) for k in meta.keys()):
                            return False

                    # match values
                    for inc in filters_v_plus:
                        if not any(inc.match(str(v)) for v in meta.values()):
                            return False
                    for exc in filters_v_minus:
                        if any(exc.match(str(v)) for v in meta.values()):
                            return False

                    # match both
                    for kp, vp in zip(filters_b_plus_keys, filters_b_plus_values):
                        good_keys = (k for k in meta.keys() if kp.match(str(k)))
                        gk_checked = 0
                        for gk in good_keys:
                            if vp.match(str(meta[gk])):
                                gk_checked += 1
                                break
                        if not gk_checked:
                            return False
                    for kp, vp in zip(filters_b_minus_keys, filters_b_minus_values):
                        good_keys = (k for k in meta.keys() if kp.match(str(k)))
                        gk_checked = 0
                        for gk in good_keys:
                            if vp.match(str(meta[gk])):
                                return False
                            gk_checked += 1
                        if not gk_checked:
                            return False

                    return True

            else:

                @optional_typecheck
                def filter_meta(meta: dict) -> bool:
                    return True

            if "filter_content" in self.cli_kwargs:
                if isinstance(self.cli_kwargs["filter_content"], str):
                    filter_content = self.cli_kwargs["filter_content"].split(",")
                else:
                    filter_content = self.cli_kwargs["filter_content"]
                assert isinstance(
                    filter_content, list
                ), f"filter_content must be a list, not {self.cli_kwargs['filter_content']}"

                # storing fast as list then in tupples for faster iteration
                filters_cont_plus = []
                filters_cont_minus = []

                for f in filter_content:
                    assert isinstance(f, str), f"Filter must be a string: '{f}'"
                    incexc = f[0]
                    assert incexc in [
                        "+",
                        "-",
                    ], f"filter 1st character must be + or -: '{f}'"
                    incexc_str = "plus" if incexc == "+" else "minus"
                    assert f[1:].strip(), f"Filter can't be an empty regex: '{f}'"
                    pattern = f[1:].strip()
                    if pattern == pattern.lower():
                        pattern = re.compile(pattern, flags=re.IGNORECASE)
                    else:
                        pattern = re.compile(pattern)
                    locals()[f"filters_cont_{incexc_str}"].append(pattern)
                filters_cont_plus = tuple(filters_cont_plus)
                filters_cont_minus = tuple(filters_cont_minus)

                @optional_typecheck
                def filter_cont(cont: str) -> bool:
                    if not all(inc.match(cont) for inc in filters_cont_plus):
                        return False
                    if any(exc.match(cont) for exc in filters_cont_minus):
                        return False
                    return True

            else:

                @optional_typecheck
                def filter_cont(cont: str) -> bool:
                    return True

            # check filtering is valid
            checked = 0
            good = 0
            ids_to_del = []
            for doc_id, doc in tqdm(
                self.loaded_embeddings.docstore._dict.items(),
                desc="Filtering",
                unit="docs",
                disable=(not is_verbose) or is_piped,
            ):
                checked += 1
                if filter_meta(doc.metadata) and filter_cont(doc.page_content):
                    good += 1
                else:
                    ids_to_del.append(doc_id)
            red(f"Keeping {good}/{checked} documents from vectorstore after filtering")
            if good == checked:
                red("Your filter matched all stored documents!")
            assert good, "No documents in the vectorstore match the given filter"

            # directly remove the filtered documents from the docstore
            # but first store the docstore before altering it to allow
            # unfiltering in the prompt
            # self.unfiltered_docstore = self.loaded_embeddings.serialize_to_bytes()
            status = self.loaded_embeddings.delete(ids_to_del)

            # checking deletions went well
            if status is False:
                raise Exception("Vectorstore filtering failed")
            elif status is None:
                raise Exception("Vectorstore filtering not implemented")
            assert len(self.loaded_embeddings.docstore._dict) == checked - len(
                ids_to_del
            ), "Something went wrong when deleting filtered out documents"
            assert len(
                self.loaded_embeddings.docstore._dict
            ), "Something went wrong when deleting filtered out documents: no document left"
            assert len(self.loaded_embeddings.docstore._dict) == len(
                self.loaded_embeddings.index_to_docstore_id
            ), "Something went wrong when deleting filtered out documents"

    @optional_typecheck
    def query_task(self, query: Optional[str]) -> dict:
        if not query:
            if self.oneoff:
                raise SystemExit("Done")
            query, self.interaction_settings = ask_user(self.interaction_settings)
        assert all(
            retriev in ["default", "multiquery", "knn", "svm", "parent"]
            for retriev in self.interaction_settings["retriever"].split("_")
        ), f"Invalid retriever value: {self.interaction_settings['retriever']}"
        retrievers = []
        if "multiquery" in self.interaction_settings["retriever"].lower():
            retrievers.append(
                create_multiquery_retriever(
                    llm=self.llm,
                    retriever=self.loaded_embeddings.as_retriever(
                        search_type="similarity_score_threshold",
                        search_kwargs={
                            "k": self.interaction_settings["top_k"],
                            "score_threshold": self.interaction_settings["relevancy"],
                        },
                    ),
                )
            )

        if "knn" in self.interaction_settings["retriever"].lower():
            retrievers.append(
                KNNRetriever.from_texts(
                    self.all_texts,
                    self.embedding_engine,
                    relevancy_threshold=self.interaction_settings["relevancy"],
                    k=self.interaction_settings["top_k"],
                )
            )
        if "svm" in self.interaction_settings["retriever"].lower():
            retrievers.append(
                SVMRetriever.from_texts(
                    self.all_texts,
                    self.embedding_engine,
                    relevancy_threshold=self.interaction_settings["relevancy"],
                    k=self.interaction_settings["top_k"],
                )
            )
        if "parent" in self.interaction_settings["retriever"].lower():
            retrievers.append(
                create_parent_retriever(
                    task=self.task,
                    loaded_embeddings=self.loaded_embeddings,
                    loaded_docs=self.loaded_docs,
                    top_k=self.interaction_settings["top_k"],
                    relevancy=self.interaction_settings["relevancy"],
                )
            )

        if "default" in self.interaction_settings["retriever"].lower():
            retrievers.append(
                self.loaded_embeddings.as_retriever(
                    search_type="similarity_score_threshold",
                    search_kwargs={
                        "k": self.interaction_settings["top_k"],
                        "score_threshold": self.interaction_settings["relevancy"],
                    },
                )
            )

        assert (
            retrievers
        ), "No retriever selected. Probably cause by a wrong cli_command or query_retrievers arg."

        if len(retrievers) == 1:
            retriever = retrievers[0]
        else:
            merge_retriever = MergerRetriever(retrievers=retrievers)

            # remove redundant results from the merged retrievers:
            filtered = EmbeddingsRedundantFilter(
                embeddings=self.embedding_engine,
                similarity_threshold=0.999,
            )
            filter_pipeline = DocumentCompressorPipeline(transformers=[filtered])
            retriever = ContextualCompressionRetriever(
                base_compressor=filter_pipeline, base_retriever=merge_retriever
            )

        if ">>>>" in query:
            sp = query.split(">>>>")
            assert (
                len(sp) == 2
            ), "The query must contain a maximum of 1 occurence of '>>>>'"
            query_fe = sp[0].strip()
            query_an = sp[1].strip()
        else:
            query_fe, query_an = copy.copy(query), copy.copy(query)
        whi(f"Query for the embeddings: {query_fe}")
        whi(f"Question to answer: {query_an}")

        # answer 0 or 1 if the document is related
        if not hasattr(self, "eval_llm"):
            failed = False
            if self.query_eval_model.backend == "openrouter":
                try:
                    self.eval_llm_params = get_supported_model_params(
                        self.query_eval_model
                    )
                except Exception as err:
                    failed = True
                    red(
                        f"Failed to get query_eval_model parameters information bypassing openrouter: '{err}'"
                    )
            if self.query_eval_model.backend != "openrouter" or failed:
                self.eval_llm_params = get_supported_model_params(self.query_eval_model)
            eval_args = {}
            if "n" in self.eval_llm_params:
                eval_args["n"] = self.query_eval_check_number
            elif self.query_eval_check_number > 1:
                red(
                    f"Model {self.query_eval_model.original} does not support parameter 'n' so will be called multiple times instead. This might cost more."
                )
                assert self.query_eval_model.backend != "openai"
            self.eval_llm = load_llm(
                modelname=self.query_eval_model,
                llm_cache=False,  # disables caching because another caching is used on top
                llm_verbosity=self.llm_verbosity,
                temperature=0 if self.query_eval_check_number == 1 else 1,
                api_base=self.llms_api_bases["query_eval_model"],
                private=self.private,
                tags=["eval_model"],
                **eval_args,
            )
            # if "anthropic" in self.query_eval_model.original.lower() or "anthropic" in self.query_eval_model.backend.lower():
            #     prompts.enable_prompt_caching("evaluate")

        # the eval doc chain needs its own caching
        if self.llm_cache:
            eval_cache_wrapper = query_eval_cache.cache
        else:

            @optional_typecheck
            def eval_cache_wrapper(func: Callable) -> Callable:
                return func

        if " object at " in self.llm._get_llm_string():
            red(
                "Found llm._get_llm_string() value that potentially "
                f"invalidates the cache: '{self.llm._get_llm_string()}'\n"
                f"Related github issue: 'https://github.com/langchain-ai/langchain/issues/23257'"
            )
        if " object at " in self.eval_llm._get_llm_string():
            red(
                "Found eval_llm._get_llm_string() value that potentially "
                f"invalidates the cache: '{self.eval_llm._get_llm_string()}'\n"
                f"Related github issue: 'https://github.com/langchain-ai/langchain/issues/23257'"
            )

        @chain
        @optional_typecheck
        def autoincrease_top_k(filtered_docs: List[Document]) -> List[Document]:
            if not self.max_top_k:
                return filtered_docs
            ratio = len(filtered_docs) / self.top_k
            if ratio >= 0.9:
                if self.top_k < self.max_top_k:
                    raise ShouldIncreaseTopKAfterLLMEvalFiltering(
                        red(
                            f"Number of documents found: {len(filtered_docs)}, "
                            f"top_k is {self.top_k} so ratio={ratio:.1f}, hence "
                            f"top_k should be increased. Max_top_k is {self.max_top_k}"
                        )
                    )
                else:
                    red(
                        f"Number of documents found: {len(filtered_docs)}, "
                        f"top_k is {self.top_k} so ratio={ratio:.1f}, hence "
                        f"top_k should be increased but we eached "
                        f"max_top_k ({self.max_top_k}) so continuing."
                    )
            return filtered_docs

        @chain
        @optional_typecheck
        @eval_cache_wrapper
        def evaluate_doc_chain(
            inputs: dict,
            query_nb: int = self.query_eval_check_number,
            eval_model_string: str = self.eval_llm._get_llm_string(),  # just for caching
            eval_prompt: str = str(prompts.evaluate.to_json()),
        ) -> List[str]:
            if isinstance(self.eval_llm, FakeListLLM):
                outputs = ["1" for i in range(self.query_eval_check_number)]
                new_p = 0
                new_c = 0

            elif "n" in self.eval_llm_params or self.query_eval_check_number == 1:
                try:
                    out = self.eval_llm._generate_with_cache(
                        prompts.evaluate.format_messages(**inputs)
                    )
                except Exception:  # retry without cache
                    out = self.eval_llm._generate(
                        prompts.evaluate.format_messages(**inputs)
                    )
                reasons = [
                    gen.generation_info["finish_reason"] for gen in out.generations
                ]
                outputs = [gen.text for gen in out.generations]
                # don't crash if finish_reason is not stop, because it can sometimes still be parsed.
                if not all(r in ["stop", "length"] for r in reasons):
                    red(
                        f"Unexpected generation finish_reason: '{reasons}' for generations: '{outputs}'"
                    )
                assert outputs, "No generations found by query eval llm"
                outputs = [parse_eval_output(o) for o in outputs]
                if out.llm_output:
                    new_p = out.llm_output["token_usage"]["prompt_tokens"]
                    new_c = out.llm_output["token_usage"]["completion_tokens"]
                else:
                    new_p = 0
                    new_c = 0

            else:
                outputs = []
                new_p = 0
                new_c = 0

                async def do_eval(subinputs):
                    try:
                        val = await self.eval_llm._agenerate_with_cache(
                            prompts.evaluate.format_messages(**subinputs)
                        )
                    except Exception:  # retrywithout cache
                        val = await self.eval_llm._agenerate(
                            prompts.evaluate.format_messages(**subinputs)
                        )
                    return val

                outs = [do_eval(inputs) for i in range(self.query_eval_check_number)]
                try:
                    loop = asyncio.get_event_loop()
                except RuntimeError:
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
                outs = loop.run_until_complete(asyncio.gather(*outs))
                for out in outs:
                    assert (
                        len(out.generations) == 1
                    ), f"Query eval llm produced more than 1 evaluations: '{out.generations}'"
                    outputs.append(out.generations[0].text)
                    finish_reason = out.generations[0].generation_info["finish_reason"]
                    if finish_reason not in ["stop", "length"]:
                        red(
                            f"Unexpected finish_reason: '{finish_reason}' for generation '{outputs[-1]}'"
                        )
                    if out.llm_output:
                        new_p += out.llm_output["token_usage"]["prompt_tokens"]
                        new_c += out.llm_output["token_usage"]["completion_tokens"]
                assert outputs, "No generations found by query eval llm"
                outputs = [parse_eval_output(o) for o in outputs]

            assert (
                len(outputs) == self.query_eval_check_number
            ), f"query eval model failed to produce {self.query_eval_check_number} outputs: '{outputs}'"

            self.eval_llm.callbacks[0].prompt_tokens += new_p
            self.eval_llm.callbacks[0].completion_tokens += new_c
            self.eval_llm.callbacks[0].total_tokens += new_p + new_c
            if self.eval_llm.callbacks[0].pbar:
                self.eval_llm.callbacks[0].pbar[-1].update(1)
            return outputs

        # uses in most places to increase concurrency limit
        multi = {"max_concurrency": WDOC_LLM_MAX_CONCURRENCY if not self.debug else 1}

        if self.task == "search":
            if self.query_eval_model is not None:
                # for some reason I needed to have at least one chain object otherwise rag_chain is a dict
                @chain
                @optional_typecheck
                def retrieve_documents(inputs):
                    return {
                        "unfiltered_docs": retriever.invoke(
                            inputs["question_for_embedding"]
                        ),
                        "question_to_answer": inputs["question_to_answer"],
                    }
                    return inputs

                meta_refilter_docs = {
                    "filtered_docs": (
                        RunnablePassthrough.assign(
                            evaluations=RunnablePassthrough.assign(
                                doc=lambda inputs: inputs["unfiltered_docs"],
                                q=lambda inputs: [
                                    inputs["question_to_answer"]
                                    for i in range(len(inputs["unfiltered_docs"]))
                                ],
                            )
                            | RunnablePassthrough.assign(
                                inputs=lambda inputs: [
                                    {"doc": d.page_content, "q": q}
                                    for d, q in zip(inputs["doc"], inputs["q"])
                                ],
                            )
                            | itemgetter("inputs")
                            | RunnableEach(
                                bound=evaluate_doc_chain.with_config(multi)
                            ).with_config(multi)
                        )
                        | refilter_docs
                        | autoincrease_top_k
                    ),
                    "unfiltered_docs": itemgetter("unfiltered_docs"),
                    "question_to_answer": itemgetter("question_to_answer"),
                }
                rag_chain = (
                    retrieve_documents
                    | sieve_documents(instance=self)
                    | pbar_chain(
                        llm=self.eval_llm,
                        len_func="len(inputs['unfiltered_docs'])",
                        desc="LLM evaluation",
                        unit="doc",
                    )
                    | meta_refilter_docs
                    | pbar_closer(llm=self.eval_llm)
                )
                tried_top_k = []
                while True:
                    if len(tried_top_k) > 10:
                        raise Exception(f"Tried more than 10 top_k: {tried_top_k}")
                    try:
                        assert self.top_k not in tried_top_k
                        tried_top_k.append(self.top_k)
                        output = rag_chain.invoke(
                            {
                                "question_for_embedding": query_fe,
                                "question_to_answer": query_an,
                            }
                        )
                        break
                    except ShouldIncreaseTopKAfterLLMEvalFiltering:
                        if self.max_top_k is None:
                            raise
                        assert (
                            self.max_top_k != self.top_k
                        ), f"Something went wrong: top_k: {self.top_k} ; max_top_k: {self.max_top_k}"
                        assert (
                            self.max_top_k > self.top_k
                        )  # if it's equal, it should have returned normally
                        new_top_k = min(int(self.top_k * 1.5), self.max_top_k)
                        assert new_top_k > self.top_k
                        assert new_top_k not in tried_top_k
                        assert new_top_k <= self.max_top_k
                        self.top_k = new_top_k

                docs = output["filtered_docs"]
            else:
                docs = retriever.invoke(query)
                if len(docs) < self.interaction_settings["top_k"]:
                    red(f"Only found {len(docs)} relevant documents")

            if self.import_mode:
                if "unfiltered_docs" in output:
                    red(
                        f"Number of documents using embeddings: {len(output['unfiltered_docs'])}"
                    )
                if "filtered_docs" in output:
                    red(
                        f"Number of documents found relevant by eval LLM: {len(output['filtered_docs'])}"
                    )
                if "relevant_filtered_docs" in output:
                    red(
                        f"Number of documents found relevant by answer LLM: {len(output['relevant_filtered_docs'])}"
                    )

                evalllmcallback = self.eval_llm.callbacks[0]
                etotal_cost = (
                    self.query_evalllm_price[0] * evalllmcallback.prompt_tokens
                    + self.query_evalllm_price[1] * evalllmcallback.completion_tokens
                )
                yel(
                    f"Tokens used by query_eval model: '{evalllmcallback.total_tokens}' (${etotal_cost:.5f})"
                )

                llmcallback = self.llm.callbacks[0]
                total_cost = (
                    self.llm_price[0] * llmcallback.prompt_tokens
                    + self.llm_price[1] * llmcallback.completion_tokens
                )
                yel(
                    f"Total tokens used by strong model: '{llmcallback.total_tokens}' (${total_cost:.5f})"
                )
                red(f"Total cost: ${total_cost + etotal_cost:.5f}")
                return output

            md_printer("\n\n# Documents")
            if WDOC_OPEN_ANKI:
                anki_nids = []
                to_print = ""
            for id, doc in enumerate(docs):
                to_print += f"## Document #{id + 1}\n"
                content = doc.page_content.strip()
                to_print += "```\n" + content + "\n ```\n"
                for k, v in doc.metadata.items():
                    to_print += f"* **{k}**: `{v}`\n"
                to_print += "\n"
                if WDOC_OPEN_ANKI and "anki_nid" in doc.metadata:
                    nid_str = str(doc.metadata["anki_nid"]).split(" ")
                    for nid in nid_str:
                        if nid not in anki_nids:
                            anki_nids.append(nid)
            md_printer(to_print)
            if self.query_eval_model is not None:
                red(
                    f"Number of documents using embeddings: {len(output['unfiltered_docs'])}"
                )
                red(
                    f"Number of documents after query eval filter: {len(output['filtered_docs'])}"
                )

            if WDOC_OPEN_ANKI and anki_nids:
                open_answ = input(
                    f"\nAnki notes found, open in anki? (yes/no/debug)\n(nids: {anki_nids})\n> "
                )
                if open_answ == "debug":
                    breakpoint()
                elif open_answ in ["y", "yes"]:
                    whi("Opening anki.")
                    query = f"nid:{','.join(anki_nids)}"
                    ankiconnect(
                        action="guiBrowse",
                        query=query,
                    )
            all_filepaths = []
            for doc in docs:
                if "path" in doc.metadata:
                    path = doc.metadata["path"]
                    try:
                        path = str(Path(path).resolve().absolute())
                    except Exception:
                        pass
                    all_filepaths.append(path)
            if all_filepaths:
                md_printer("### All file paths")
                md_printer("* " + "\n* ".join(all_filepaths))

            evalllmcallback = self.eval_llm.callbacks[0]
            etotal_cost = (
                self.query_evalllm_price[0] * evalllmcallback.prompt_tokens
                + self.query_evalllm_price[1] * evalllmcallback.completion_tokens
            )
            yel(
                f"Tokens used by query_eval model: '{evalllmcallback.total_tokens}' (${etotal_cost:.5f})"
            )

            red(f"Total cost: ${etotal_cost:.5f}")
            self.latest_cost = etotal_cost

        else:
            # for some reason I needed to have at least one chain object otherwise rag_chain is a dict
            @chain
            @optional_typecheck
            def retrieve_documents(inputs):
                return {
                    "unfiltered_docs": retriever.invoke(
                        inputs["question_for_embedding"]
                    ),
                    "question_to_answer": inputs["question_to_answer"],
                }
                return inputs

            meta_refilter_docs = {
                "filtered_docs": (
                    RunnablePassthrough.assign(
                        evaluations=RunnablePassthrough.assign(
                            doc=lambda inputs: inputs["unfiltered_docs"],
                            q=lambda inputs: [
                                inputs["question_to_answer"]
                                for i in range(len(inputs["unfiltered_docs"]))
                            ],
                        )
                        | RunnablePassthrough.assign(
                            inputs=lambda inputs: [
                                {"doc": d.page_content, "q": q}
                                for d, q in zip(inputs["doc"], inputs["q"])
                            ]
                        )
                        | itemgetter("inputs")
                        | RunnableEach(
                            bound=evaluate_doc_chain.with_config(multi)
                        ).with_config(multi)
                    )
                    | refilter_docs
                    | autoincrease_top_k
                ),
                "unfiltered_docs": itemgetter("unfiltered_docs"),
                "question_to_answer": itemgetter("question_to_answer"),
            }
            answer_each_doc_chain = (
                prompts.answer | self.llm.bind(max_tokens=1000) | StrOutputParser()
            )

            answer_all_docs = RunnablePassthrough.assign(
                inputs=lambda inputs: [
                    {"context": d.page_content, "question_to_answer": q}
                    for d, q in zip(
                        inputs["filtered_docs"],
                        [inputs["question_to_answer"]] * len(inputs["filtered_docs"]),
                    )
                ]
            ) | {
                "intermediate_answers": itemgetter("inputs")
                | RunnableEach(bound=answer_each_doc_chain),
                "question_to_answer": itemgetter("question_to_answer"),
                "filtered_docs": itemgetter("filtered_docs"),
                "unfiltered_docs": itemgetter("unfiltered_docs"),
            }

            rag_chain = (
                retrieve_documents
                | sieve_documents(instance=self)
                | pbar_chain(
                    llm=self.eval_llm,
                    len_func="len(inputs['unfiltered_docs'])",
                    desc="LLM evaluation",
                    unit="doc",
                )
                | meta_refilter_docs
                | pbar_closer(llm=self.eval_llm)
                | pbar_chain(
                    llm=self.llm,
                    len_func="len(inputs['filtered_docs'])",
                    desc="Answering each",
                    unit="doc",
                )
                | answer_all_docs
                | pbar_closer(llm=self.llm)
            )

            if is_verbose:
                rag_chain.get_graph().print_ascii()

            chain_time = 0
            start_time = time.time()
            tried_top_k = []
            while True:
                if len(tried_top_k) > 10:
                    raise Exception(f"Tried more than 10 top_k: {tried_top_k}")
                try:
                    assert self.top_k not in tried_top_k
                    tried_top_k.append(self.top_k)
                    output = rag_chain.invoke(
                        {
                            "question_for_embedding": query_fe,
                            "question_to_answer": query_an,
                        }
                    )
                    break
                except ShouldIncreaseTopKAfterLLMEvalFiltering:
                    if self.max_top_k is None:
                        raise
                    elif self.max_top_k == self.top_k:
                        break
                    assert self.max_top_k > self.top_k
                    new_top_k = min(int(self.top_k * 1.5), self.max_top_k)
                    assert new_top_k > self.top_k
                    assert new_top_k not in tried_top_k
                    assert new_top_k <= self.max_top_k
                    self.top_k = new_top_k
                except NoDocumentsRetrieved:
                    return {
                        "error": md_printer(
                            f"## No documents were retrieved with query '{query_fe}'",
                            color="red",
                        )
                    }
                except NoDocumentsAfterLLMEvalFiltering:
                    return {
                        "error": md_printer(
                            f"## No documents remained after query eval LLM filtering using question '{query_an}'",
                            color="red",
                        )
                    }
            chain_time = time.time() - start_time

            assert len(output["intermediate_answers"]) == len(output["filtered_docs"])

            if len(output["intermediate_answers"]) > 1:
                # next step is to combine the intermediate answers into a single answer
                final_answer_chain = RunnablePassthrough.assign(
                    final_answer=RunnablePassthrough.assign(
                        question=lambda inputs: inputs["question_to_answer"],
                        intermediate_answers=lambda inputs: collate_intermediate_answers(
                            list_ia=inputs["intermediate_answers"],
                        ),
                    )
                    | prompts.combine
                    | self.llm
                    | StrOutputParser()
                )

                # add the document hash as source to each intermediate answer, they will then be combined together and replaced again last minute by more legible identifiers
                for ifd, fd in enumerate(output["filtered_docs"]):
                    ia = output["intermediate_answers"][ifd]
                    doc_hash = fd.metadata["content_hash"][:5]
                    output["intermediate_answers"][
                        ifd
                    ] = f"Source identifier: [{doc_hash}]\n{ia}"
                source_hashes = {
                    d.metadata["content_hash"][:5]: str(idoc + 1)
                    for idoc, d in enumerate(output["filtered_docs"])
                }

                @optional_typecheck
                def source_replace(input: str) -> str:
                    for h, idoc in source_hashes.items():
                        input = input.replace(h, idoc)
                    return input

                llmcallback = self.llm.callbacks[0]
                cost_before_combine = (
                    self.llm_price[0] * llmcallback.prompt_tokens
                    + self.llm_price[1] * llmcallback.completion_tokens
                )

                # group the intermediate answers by batch, then do a batch reduce mapping
                # each batch is at least 2 intermediate answers and maxes at
                # batch_tkn_size tokens to avoid losing anything because of
                # the context
                all_intermediate_answers = [output["intermediate_answers"]]
                pbar = tqdm(
                    desc="Combining answers",
                    unit="answer",
                    total=len(output["intermediate_answers"]),
                    # disable=not is_verbose,
                    disable=is_piped,
                )
                temp_interm_answ = output["intermediate_answers"]
                temp_interm_answ = thinking_answer_parser(temp_interm_answ)["answer"]
                while True:
                    batches = [[]]
                    # disregard IRRELEVANT answers
                    temp_interm_answ = [
                        ia for ia in temp_interm_answ if check_intermediate_answer(ia)
                    ]
                    batches = semantic_batching(temp_interm_answ, self.embedding_engine)
                    batch_args = [
                        {"question_to_answer": query_an, "intermediate_answers": b}
                        for b in batches
                    ]
                    temp_interm_answ = []
                    batch_result = final_answer_chain.batch(batch_args)
                    n_trial = 3
                    for ia, a in enumerate(batch_result):
                        trial = 0
                        for trial in range(n_trial):
                            try:
                                o = thinking_answer_parser(
                                    a["final_answer"],
                                    strict=True,
                                )["answer"]
                                temp_interm_answ.append(o)
                                break
                            except Exception as e:
                                red(
                                    f"Error at trial {trial} when separating "
                                    "thinking from answer from LLM output.\n\n"
                                    f"The full answer is: '{a}'\n\n"
                                    f"The error was: '{e}'\n\n"
                                    "Retrying "
                                    "this specific batch to make sure we don't"
                                    " loose intermediate answers"
                                )
                                # modify the batch slightly to bypass the cache
                                altered_batch = batch_args[ia]
                                altered_batch["question_to_answer"] += "."
                                altered_batch["intermediate_answers"] += "."
                                a = final_answer_chain.batch([altered_batch])[0]

                    all_intermediate_answers.append(temp_interm_answ)
                    pbar.n = pbar.total - len(temp_interm_answ) + 1
                    pbar.update(0)
                    if len(temp_interm_answ) == 1:
                        break

                assert pbar.n == pbar.total
                pbar.close()
                assert len(all_intermediate_answers[-1]) == 1

                final_answer = all_intermediate_answers[-1][0]
                output["all_intermediate_answers"] = all_intermediate_answers
            else:
                final_answer = output["intermediate_answers"][0]
                output["all_intermediate_answers"] = [final_answer]

                def source_replace(input):
                    return input

                all_intermediate_answers = [final_answer]

            # prepare the content of the output
            output["final_answer"] = final_answer
            output["relevant_filtered_docs"] = []
            output["relevant_intermediate_answers"] = []
            for ia, a in enumerate(output["intermediate_answers"]):
                if check_intermediate_answer(a):
                    output["relevant_filtered_docs"].append(output["filtered_docs"][ia])
                    output["relevant_intermediate_answers"].append(a)

            # display sources (i.e. documents used to answer)
            md_printer("---")
            if not output["relevant_intermediate_answers"]:
                md_printer(
                    "\n\n# No document filtered so no intermediate answers to combine.\nThe answer will be based purely on the LLM's internal knowledge.",
                    color="red",
                )
                md_printer(
                    "\n\n# No document filtered so no intermediate answers to combine"
                )
            else:
                md_printer("\n\n# Intermediate answers for each document:")
            for counter, (ia, doc) in enumerate(
                zip(
                    output["relevant_intermediate_answers"],
                    output["relevant_filtered_docs"],
                )
            ):
                to_print = f"## Document #{counter + 1}\n"
                content = doc.page_content.strip()
                to_print += "```\n" + content + "\n ```\n"
                for k, v in doc.metadata.items():
                    to_print += f"* **{k}**: `{v}`\n"
                ia = thinking_answer_parser(ia)
                # print either both thinking and answer or just answer
                # ia = "### Thinking:\n" + ia["thinking"] + "\n\n" + "### Answer:\n" + ia["answer"]
                ia = ia["answer"]
                to_print += indent("### Intermediate answer:\n" + ia, "> ")
                md_printer(source_replace(to_print))

            # print the final answer
            fa = thinking_answer_parser(output["final_answer"])
            # fa = "### Thinking:\n" + fa["thinking"] + "\n\n" + "### Answer:\n" + fa["answer"]
            fa = fa["answer"]
            md_printer("---")
            md_printer(indent(f"# Answer:\n{source_replace(fa)}\n", "> "))

            # print the breakdown of documents used and chain time
            red(
                f"Number of documents using embeddings: {len(output['unfiltered_docs'])}"
            )
            red(
                f"Number of documents found relevant by eval LLM: {len(output['filtered_docs'])}"
            )
            red(
                f"Number of documents found relevant by answer LLM: {len(output['relevant_filtered_docs'])}"
            )
            if len(all_intermediate_answers) > 1:
                extra = "->".join([str(len(ia)) for ia in all_intermediate_answers])
                extra = f"({extra})"
            else:
                extra = ""
            red(
                f"Number of steps to combine intermediate answers: {len(all_intermediate_answers) - 1} {extra}"
            )
            red(f"Time took by the chain: {chain_time:.2f}s")

            evalllmcallback = self.eval_llm.callbacks[0]
            etotal_cost = (
                self.query_evalllm_price[0] * evalllmcallback.prompt_tokens
                + self.query_evalllm_price[1] * evalllmcallback.completion_tokens
            )
            yel(
                f"Tokens used by query_eval model: '{evalllmcallback.total_tokens}' (${etotal_cost:.5f})"
            )

            llmcallback = self.llm.callbacks[0]
            total_cost = (
                self.llm_price[0] * llmcallback.prompt_tokens
                + self.llm_price[1] * llmcallback.completion_tokens
            )
            yel(
                f"Total tokens used by strong model: '{llmcallback.total_tokens}' (${total_cost:.5f})"
            )
            if "cost_before_combine" in locals():
                combine_cost = total_cost - cost_before_combine
                yel(
                    f"Tokens used by strong model to combine the intermediate answers: ${combine_cost:.5f}"
                )

            red(f"Total cost: ${total_cost + etotal_cost:.5f}")

            self.latest_cost = total_cost + etotal_cost

            return output

    @staticmethod
    @optional_typecheck
    @set_parse_file_help_md_as_docstring
    def parse_file(
        path: Optional[Union[str, Path]] = None,
        filetype: str = "auto",
        format: str = "text",
        cli_kwargs: Optional[dict] = None,
        debug: bool = False,
        verbose: bool = False,
        **kwargs,
    ) -> Union[List[Document], str, List[dict]]:
        """
        This docstring is dynamically updated with the content of wdoc/docs/parse_file_help.md
        """
        assert format in [
            "text",
            "xml",
            "langchain",
            "langchain_dict",
        ], f"Unexpected --format value: '{format}'"
        default_cli_kwargs = {
            "llm_name": ModelName("testing/testing"),
            "task": "query",
            "backend": "loky",  # doesn't matter because n_jobs is 1 anyway
            "n_jobs": 1,
            "loading_failure": "crash",
        }

        if debug:
            debug_exceptions()

        if cli_kwargs is not None:
            default_cli_kwargs.update(cli_kwargs)

        assert (
            "task" not in kwargs
        ), "Cannot give --task argument if we are only parsing"

        if kwargs:
            kwargs = DocDict(kwargs)

        # all loaders need a path arg except anki
        if filetype == "anki" and path:
            red(
                "You supplied a --path argument even though the filetype is `anki`, we must ignore `path` in that case."
            )
        else:
            if not path:
                red(
                    "You did not specify a --path argument, this will probably cause issues."
                )
            else:
                kwargs["path"] = path

        out = batch_load_doc(
            filetype=filetype,
            **default_cli_kwargs,
            **kwargs,
        )
        if format == "text":
            n = len(out)
            return (
                "Parsed documents\n"
                + "\n".join(
                    [f"Doc #{i}/{n}\n{d.page_content}\n\n" for i, d in enumerate(out)]
                ).rstrip()
            )
        elif format == "xml":
            return (
                "<documents>\n"
                + "\n".join(
                    [
                        f"<doc id={i}>\n{d.page_content}\n</doc>"
                        for i, d in enumerate(out)
                    ]
                )
                + "\n</documents>"
            )
        elif format == "langchain":
            return out
        elif format == "langchain_dict":
            return [
                {"page_content": doc.page_content, "metadata": doc.metadata}
                for doc in out
            ]
        else:
            raise ValueError(format)


def debug_exceptions(instance: Optional[wdoc] = None) -> None:
    "open a debugger is --debug is set"

    def handle_exception(exc_type, exc_value, exc_traceback):
        if not issubclass(exc_type, KeyboardInterrupt):

            @optional_typecheck
            def p(message: str) -> None:
                "print error, in red if possible"
                if instance:
                    red(instance.ntfy(message))
                else:
                    try:
                        red(message)
                    except Exception:
                        print(message)

            p(
                "\n--verbose was used so opening debug console at the "
                "appropriate frame. Press 'c' to continue to the frame "
                "of this print."
            )
            p(
                "Please open an issue on github and include the trace. It's "
                "tremendously useful for me as there are many small bugs that "
                "can be quickly squashed if users just told me about it :)"
            )
            [p(line) for line in traceback.format_tb(exc_traceback)]
            p(str(exc_type) + " : " + str(exc_value))
            if hasattr(exc_value, "__cause__") and hasattr(
                exc_value.__cause__, "__traceback__"
            ):
                p("Detected a cause to the exception, opening the cause first")
                pdb.post_mortem(exc_value.__cause__.__traceback__)
                p("Out of the __cause__, now debugging the higher traceback:")
            pdb.post_mortem(exc_traceback)
            p("You are now in the exception handling frame.")
            breakpoint()
            sys.exit(1)

    sys.excepthook = handle_exception
    faulthandler.enable()


if not is_private:
    create_langfuse_callback(f"wdoc_{wdoc.VERSION}")



"""
Exception classes
"""


class NoDocumentsRetrieved(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class NoDocumentsAfterLLMEvalFiltering(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ShouldIncreaseTopKAfterLLMEvalFiltering(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidDocEvaluationByLLMEval(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class UnexpectedDocDictArgument(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(
            f"You're trying to use an argument for a filetype that does not expect it: {message}"
        )


class TimeoutPdfLoaderError(Exception):
    def __init__(self) -> None:
        super().__init__()



"""
TheFiche: A module for generating and exporting Logseq pages based on wdoc queries.

This module provides functionality to create structured Logseq pages
with content generated from wdoc queries, including metadata and properties.
"""

import inspect
import json
import re
import time
from datetime import datetime
from pathlib import Path, PosixPath
from textwrap import indent
from typing import Callable, Dict, List, Optional, Tuple, Union

import fire
import litellm
from beartype import beartype
from joblib import Memory
from LogseqMarkdownParser import LogseqBlock, LogseqPage, parse_file, parse_text
from loguru import logger

from wdoc import wdoc

VERSION = "1.3"

# logger
logger.add(
    "logs.txt",
    rotation="100MB",
    retention=5,
    format="{time} {level} {thread} TheFiche {process} {function} {line} {message}",
    level="DEBUG",
    enqueue=True,
    colorize=False,
)


def p(*args):
    "simple way to log to logs.txt and to print"
    print(*args)
    logger.info(*args)


d = datetime.today()
today = f"{d.day:02d}/{d.month:02d}/{d.year:04d}"

mem = Memory(".cache", verbose=False)


def chat(
    model: str,
    messages: List[Dict],
    temperature: float,
    **kwargs: Dict,
) -> Dict:
    """call to an LLM api. Cached."""
    answer = litellm.completion(
        model=model,
        messages=messages,
        temperature=temperature,
        stream=False,
        **kwargs,
    ).json()
    assert all(
        a["finish_reason"] == "stop" for a in answer["choices"]
    ), f"Found bad finish_reason: '{answer}'"
    return answer


def run_wdoc(query: str, kwargs2: dict) -> Tuple[wdoc, dict]:
    "call to wdoc, optionaly cached"
    instance = wdoc(
        task="query",
        import_mode=True,
        query=query,
        **kwargs2,
    )
    fiche = instance.query_task(query=query)
    assert "error" not in fiche

    if len(fiche["all_intermediate_answers"]) > 1:
        extra = "->".join([str(len(ia)) for ia in fiche["all_intermediate_answers"]])
        extra = f"({extra})"
    else:
        extra = ""

    props = {
        "block_type": "wdoc_the_fiche",
        "wdoc_version": instance.VERSION,
        "wdoc_model": f"{instance.modelname} of {instance.modelbackend}",
        "wdoc_evalmodel": f"{instance.query_eval_modelname} of {instance.query_eval_modelbackend}",
        "wdoc_evalmodel_check_nb": instance.query_eval_check_number,
        "wdoc_cost": f"{float(instance.latest_cost):.5f}",
        "wdoc_n_docs_found": len(fiche["unfiltered_docs"]),
        "wdoc_n_docs_filtered": len(fiche["filtered_docs"]),
        "wdoc_n_docs_used": len(fiche["relevant_filtered_docs"]),
        "wdoc_n_combine_steps": str(len(fiche["all_intermediate_answers"]))
        + " "
        + extra,
        "wdoc_kwargs": json.dumps(kwargs2, ensure_ascii=False),
        "the_fiche_version": VERSION,
        "the_fiche_date": today,
        "the_fiche_timestamp": int(time.time()),
        "the_fiche_query": query,
    }
    return fiche, props


@beartype
class TheFiche:
    """
    A class for generating and exporting Logseq pages based on wdoc queries.

    This class encapsulates the process of creating a Logseq page with content
    generated from a wdoc query, including metadata and properties.
    """

    def __init__(
        self,
        query: str,
        logseq_page: Union[str, PosixPath],
        overwrite: bool = False,
        sources_location: str = "as_pages",
        sources_ref_as_prop: bool = False,
        use_cache: bool = True,
        logseq_linkify: bool = True,
        the_fiche_callback: Optional[Callable] = None,
        **kwargs,
    ):
        """
        Initialize a TheFiche instance and generate a Logseq page.

        Args:
            query (str): The query to be processed by wdoc.
            logseq_page (Union[str, PosixPath]): The path to the Logseq page file.
            overwrite (bool, optional): Whether to overwrite an existing file. Defaults to False. If False, will append to the file instead of overwriting. Else, will also overwrite sources if present.
            sources_location (str): If 'as_pages', will store each source as its own page in a 'TheFiche___' namespace. If 'below', sources will be written at the end of the page. Default to "as_pages".
            sources_ref_as_prop (bool): if True, make sure the sources appear as block properties instead of leaving them as is. Default to False.
            use_cache (bool): set to False to bypass the cache, default True.
            logseq_linkify (bool): If True, will ask wdoc's strong LLM to find the import keywords and automatically replace them in the output file by logseq [[links]], enabling the use of graph properties. Default to True.
            the_fiche_callback (callable): will be called at the end of the run with as argument all what's in locals(). Use only if you know what you're doing.
            **kwargs: Additional keyword arguments to pass to wdoc.

        Raises:
            AssertionError: If the file exists and overwrite is False, or if the ratio of used/found documents is too high.
        """
        p(
            f"Starting TheFiche with args: {query}, {logseq_page}, {overwrite}, {sources_location} and kwargs: {kwargs}"
        )
        assert sources_location in ["as_pages", "below"]
        logseq_page = Path(logseq_page)

        all_kwargs = kwargs.copy()

        # posixpath can't be serialized to json
        for k, v in all_kwargs.items():
            if isinstance(v, PosixPath):
                all_kwargs[k] = str(v)

        if use_cache:
            cached = mem.cache(run_wdoc)
        else:
            cached = run_wdoc

        fiche, props = cached(query=query, kwargs2=all_kwargs)

        p(f"Fiche properties: {props}")

        n_used = props["wdoc_n_docs_used"]
        n_found = props["wdoc_n_docs_found"]

        text = fiche["final_answer"]
        assert text.strip(), fiche

        # make sure that md line that start with an int are prepended with
        # a '-' ; I know this is not standard markdown but it avoid issues
        # with logseq
        lines = text.splitlines(keepends=True)
        for il, li in enumerate(lines):
            ls = li.lstrip()
            if not ls:
                continue
            first = ls[0]
            if first.isdigit():
                lines[il] = li.replace(first, "- " + first, 1)
        text = "".join(lines)

        # also, any line starting with > must be escaped to not be seen as a quote
        lines = text.splitlines(keepends=True)
        for il, li in enumerate(lines):
            ls = li.lstrip()
            ls = ls.replace("-", "", 1).lstrip()
            if not ls:
                continue
            if ls.startswith(">"):
                lines[il] = li.replace(">", r"\>", 1)
        text = "".join(lines)

        # prepare sources hash dict
        doc_hash = {
            d.metadata["content_hash"][:5]: d.metadata for d in fiche["filtered_docs"]
        }
        # discard sources that are not used
        used_hash = []
        for d, dm in doc_hash.items():
            if d in text:
                used_hash.append(d)
            # note: we replace content_hash by all_hash to avoid collisions, as sources are originally refered only by the first 5 letters of content_hash, we now use 5 of all_hash
            new_h = dm["all_hash"][:5]
            text = text.replace(d, f" [[{new_h}]] ")
        text = text.replace(" , ", ", ")
        if not used_hash:
            p("No documents seem to be sourced")

        if logseq_linkify:
            if use_cache:
                cached_chat = mem.cache(chat)
            else:
                cached_chat = chat
            litellm.drop_params = True
            # remove the sources from the text
            text_wo_s = text
            for dh, dm in doc_hash.items():
                new_h = dm["all_hash"][:5]
                assert dh not in text_wo_s
                if new_h in text_wo_s:
                    assert f"[[{new_h}]]" in text_wo_s
                    text_wo_s = re.sub(
                        f"[, ]*\[\[{new_h}\]\][, ]*",
                        "",
                        text_wo_s,
                    )
                    assert f"[[{new_h}]]" not in text_wo_s
                    assert new_h not in text_wo_s
                    text_wo_s = text_wo_s.replace(" []", "")
                    text_wo_s = text_wo_s.replace("[]", "")
            messages = [
                {
                    "role": "system",
                    "content": """Your task is, given a text, to reply a newline separated list of tags while following some rules.
The purpose of this task is to identify the strings in my text that I could use to create graph like links in my corpus of texts.

Rules:
'''
- The tags can contain multiple words.
- The tags have to be exclusive, for example "recette de cuisine" is bad because you can use instead "cuisine".
- The tags have to be as atomic as possible, for example instead of "HER2 gene" use "HER2".
- Use the singular form of the tag, for exampel not "ulcères" but "ulcère".
- The tags ALWAYS have to exist in the text as is. This is the most important rule. You have to respect the accents, the letters used, etc. Even if it's misspelled!
- The tags have to be about details and specifics, not really big picture stuff. Think for example which drugs are mentionned in this lesson.
- Write each keyword in its own line, so newline separated.
- Take a deep breath before answering.
- Wrap your tags list in a <tags> banner.
- Use as many tags as you deem fit. An absolute minimum of 1 every 3 lines seems about right.
- I won't accept a tag that is not a substring of the whole text. IT HAS TO APPEAR IN THE TEXT AS IS.
'''


Example of an appropriate answer on a text about hemostasis:
'''
Okay, I've read the whole text. I'm now taking a deep breath.
.
.
Okay here's my answer.

<tags>
hemostase
ADAMTS13
VWF
facteur von willebrand
microangiopathie thrombotique
PTT
purpura thrombotique thrombopenique
purpura
SHU
syndrome hemolytique et uremique
infection
anticorps
</tags>
'''
""",
                },
                {"role": "user", "content": text_wo_s},
            ]
            tag_answer = cached_chat(
                messages=messages,
                model=(
                    inspect.signature(wdoc).parameters["modelname"].default
                    if "modelname" not in kwargs
                    else kwargs["modelname"]
                ),
                temperature=0,
            )
            tags_text = tag_answer["choices"][0]["message"]["content"]
            assert (
                "<tags>" in tags_text and "</tags>" in tags_text
            ), f"missing <tags> or </tags> in new text:\n'''\n{tags_text}\n'''"
            tags = tags_text.split("<tags>", 1)[1].split("</tags>", 1)[0].splitlines()
            tags = [t.strip() for t in tags if t.strip()]
            for t in tags:
                assert "[[" not in t and "]]" not in t, f"tag contains brackets: {t}"
            tags = sorted(tags, key=len)
            assert tags, f"No tags found in tags_text:\n'''\n{tags_text}\n'''"
            text_wo_tags = text

            # tags can be part of another tag
            def find_substrings(string_list: List[str]) -> List[str]:
                result = []
                for i, s1 in enumerate(string_list):
                    for j, s2 in enumerate(string_list):
                        if i != j and s1 in s2:
                            result.append(s1)
                            break
                result = list(set(result))
                assert len(result) <= len(string_list)
                return result

            sub_tags = find_substrings(tags)
            sup_tags = [
                t
                for t in tags
                if t not in sub_tags and re.search(t, text, re.IGNORECASE)
            ]
            if sup_tags:
                for it, t in enumerate(sup_tags):
                    tb = r"\b" + t + r"\b"
                    if not re.search(tb, text, re.IGNORECASE):
                        p(f"Couldn't find tag {t} in text")
                    else:
                        p(f"Found tag {t} in text")
                        text = re.sub(tb, "[[" + t + "]]", text, re.IGNORECASE)
            if sub_tags:
                assert sup_tags
                lines = text.splitlines(keepends=True)
                new_text = ""
                for line in lines:
                    for t in sub_tags:
                        t = t.strip()
                        assert t
                        assert "[" not in t and "]" not in t
                        tb = r"\b" + t + r"\b"
                        if not re.search(tb, line, re.IGNORECASE):
                            continue

                        if line == line.rstrip():
                            line = line + " [[" + t + "]]"
                        else:
                            head = line.rstrip()
                            tail = line[len(head) :]
                            assert head == head.rstrip()
                            assert tail != tail.rstrip(), tail
                            line = head + " [[" + t + "]]" + tail
                            assert line != line.rstrip()

                    new_text += line
                # assert text != new_text
                if text == new_text:
                    p(
                        f"Text was not different after checking for subtags, which are: '{sub_tags}'"
                    )

                text = new_text

        # create logseq page object but don't save it yet
        if not text.startswith("- "):
            text = "- " + text
        # fix logseq indentation
        if True:  # or text.count("    ") % text.count("  ") != 0:
            text = text.replace("  ", "    ")
        content = parse_text(text)
        content.page_properties.update(props)
        assert content.page_properties

        # add documents source
        if sources_location == "below":
            content.blocks.append(LogseqBlock("- # Sources\n  collapsed:: true"))
            for dh, dm in doc_hash.items():
                if dh not in used_hash:
                    continue
                cont = [
                    fd.page_content
                    for fd in fiche["filtered_docs"]
                    if fd.metadata["all_hash"] == dm["all_hash"]
                ]
                assert (
                    len(cont) == 1
                ), f"Found multiple sources with the same hash! {cont}"
                cont = cont[0].strip()
                new_block = LogseqBlock(f"- [[{dh}]]: {indent(cont, '  ').strip()}")
                for k, v in dm.items():
                    new_block.set_property(k, v)
                diff = 4 - new_block.indentation_level % 4
                if diff != 0:
                    diff += 4
                new_block.indentation_level += 4 + diff
                content.blocks.append(new_block)

        elif sources_location == "as_pages":
            logseq_dir = logseq_page.parent
            for dh, dm in doc_hash.items():
                if dh not in used_hash:
                    continue
                source_path = logseq_dir / (
                    "TheFicheSources___" + dm["all_hash"] + ".md"
                )
                cont = [
                    fd.page_content
                    for fd in fiche["filtered_docs"]
                    if fd.metadata["all_hash"] == dm["all_hash"]
                ]
                assert (
                    len(cont) == 1
                ), f"Found multiple sources with the same hash! {cont}"
                cont = cont[0].strip()
                new_h = dm["all_hash"][:5]

                if overwrite:
                    source_path.unlink(missing_ok=True)

                conflict = False
                if source_path.exists():
                    p(
                        f"Warning: a source for {dh} ({new_h}) already exists at {source_path}"
                    )
                    prev_source = parse_file(source_path)
                    if prev_source.page_properties["content_hash"][:5].startswith(dh):
                        if not prev_source.page_properties["all_hash"].startswith(
                            new_h
                        ):
                            p(
                                f"Found previous source with the same name and overlapping content but different alias: page: {source_path}, dh: {dh}, new_h: {new_h}\nCreating a conflict file"
                            )
                            conflict = True
                    else:
                        p(
                            f"Found previous source with the same name but does not contain the new content: {source_path}, dh: {dh}, new_h: {new_h}\nCreating a conflict file"
                        )
                        conflict = True

                if conflict:
                    sp = source_path.parent / (source_path.name + "_conflict")
                else:
                    sp = source_path
                logger.info(f"Creating source page for {dh} ({new_h}) at {sp}")
                cont = indent(cont, "  ").strip()
                if not cont.startswith("- "):
                    cont = "- " + cont
                source_page = parse_text(cont)
                source_page.page_properties.update(dm.copy())
                source_page.page_properties["alias"] = new_h
                source_page.export_to(
                    file_path=sp.absolute(),
                    overwrite=True,
                    allow_empty=False,
                )
        else:
            raise ValueError(sources_location)

        if sources_ref_as_prop:
            # make it so that the sources appear as block properties instead of in the content
            for ib, b in enumerate(content.blocks):
                for dh, dm in doc_hash.items():
                    new_h = dm["all_hash"][:5]
                    assert dh not in b.content
                    if new_h in b.content:
                        assert f"[[{new_h}]]" in b.content
                        content.blocks[ib].content = re.sub(
                            f"[, ]*\[\[{new_h}\]\][, ]*",
                            "",
                            b.content,
                        )
                        assert f"[[{new_h}]]" not in b.content
                        assert new_h not in content.blocks[ib].content
                        content.blocks[ib].content = content.blocks[ib].content.replace(
                            " []", ""
                        )
                        content.blocks[ib].content = content.blocks[ib].content.replace(
                            "[]", ""
                        )
                        content.blocks[ib].set_property("source", f"[[{new_h}]]")

        # save to file
        if (not logseq_page.absolute().exists()) or overwrite:
            content.export_to(
                file_path=logseq_page.absolute(),
                overwrite=False if not overwrite else True,
                allow_empty=False,
            )
        else:
            prev_content = parse_file(logseq_page)
            prev_content.blocks.append(LogseqBlock("- ---"))

            new_block = LogseqBlock(f"- # {today}")
            for k, v in content.page_properties.items():
                new_block.set_property(k, v)

            prev_content.blocks.append(new_block)
            # to over indent the new blocks
            # for block in content.blocks:
            #     diff = (4 - block.indentation_level % 4)
            #     if diff != 0:
            #         diff += 4
            #     block.indentation_level += 4 + diff
            #     prev_content.blocks.append(block)
            prev_content.export_to(
                file_path=logseq_page.absolute(),
                overwrite=True,
                allow_empty=False,
            )

        if the_fiche_callback:
            p(f"Using callback '{the_fiche_callback}'")
            the_fiche_callback(**locals())

        p("Done")


if __name__ == "__main__":
    fire.Fire(TheFiche)



# needed to fix ollama 'event loop closed' error
# thanks to https://github.com/BerriAI/litellm/pull/7625/files
try:
    import nest_asyncio

    nest_asyncio.apply()
except Exception as e:
    print(f"Failed to patch asyncio loop using nest_asyncio. Error: '{e}'")

from . import utils
from .wdoc import wdoc

__VERSION__ = wdoc.VERSION


__all__ = ["wdoc", "utils"]



"""
Entry point file
"""

import json
import os
import re
import sys
from pathlib import Path

import fire

# Just in case: we set the env variable of WDOC_PRIVATE_MODE as early as possible, even before importing wdoc
if re.findall(r"\b--private\b", " ".join(sys.argv)):
    print("Detected --private mode: setting WDOC_PRIVATE_MODE to True")
    os.environ["WDOC_PRIVATE_MODE"] = True

from .wdoc import is_verbose, wdoc, whi


def cli_launcher() -> None:
    """entry point function, modifies arguments on the fly for easier
    shorthands then call wdoc"""
    sysline = " ".join(sys.argv)

    # turn 'wdoc parse' into 'wdoc_parse_file'
    if (
        "wdoc parse " in sysline
        or "wdoc parse_file " in sysline
        or "__main__.py parse " in sysline
        or "__main__.py parse_file  " in sysline
    ):
        if is_verbose:
            whi("Replacing 'wdoc parse' by 'wdoc_parse_file'")
        sys.argv[0] = str(Path(sys.argv[0]).parent / "wdoc_parse_file")
        del sys.argv[1]
        sysline = " ".join(sys.argv)
        cli_parse_file()
        raise SystemExit(0)

    if " --version" in sysline:
        print(f"wdoc version: {wdoc.VERSION}")
        raise SystemExit(0)
    elif " --help" in sysline or " -h" in sysline:
        print("Showing help")
        wdoc.md_printer(wdoc.__doc__)
        raise SystemExit(0)
    elif " --completion" in sysline:
        if " -- --completion" in sysline:
            fire.Fire(wdoc)
            raise SystemExit(0)
        else:
            raise Exception(
                "To create completion scripts, use '-- --completion' as arguments."
            )
    elif len(sys.argv) == 1:
        whi("No args shown. Use '--help' to display the help.")
        raise SystemExit(0)

    # while we're at it, make it so that
    # "wdoc summary" is parsed like "wdoc --task=summary"
    if " --task" not in sysline:
        arg_replacement_rules = {
            "query": "--task=query",
            "search": "--task=search",
            "summarize": "--task=summarize",
            "summarize_then_query": "--task=summarize_then_query",
            "summary": "--task=summarize",
            "summary_then_query": "--task=summarize_then_query",
        }
        if sys.argv[1] in arg_replacement_rules:
            bef = sys.argv[1]
            aft = arg_replacement_rules[bef]
            if is_verbose:
                whi(f"Replaced argument '{bef}' to '{aft}'")
            sys.argv[1] = aft

    # make it so that 'wdoc --task=query THING' becomes 'wdoc --task=query --path=THING'
    if (
        ("--path" not in sysline)
        and (not re.findall(r"\bstring\b", sysline))
        and (not re.findall(r"\banki\b", sysline))
    ):
        # if string is present that can be because of --filetype=string or
        # --filetype=anki, in which case 'path' argument is not needed
        path = sys.argv[2]
        newarg = f"--path={path}"
        sys.argv[2] = newarg
        if is_verbose:
            whi(f"Replaced '{path}' to '{newarg}'")

    fire.Fire(wdoc)


def cli_parse_file() -> None:
    sys_args = sys.argv
    if "--help" in sys_args:
        print("Showing help")
        wdoc.md_printer(wdoc.parse_file.__doc__)
        # help(wdoc.parse_file)
        # print(wdoc.parse_file.__doc__)
        # fire.Fire(wdoc.parse_file)
        raise SystemExit(0)

    if "--pipe" in sys_args or "pipe" in sys_args:
        # parse args manually to allow piping
        args = []
        kwargs = {}
        for val in sys_args:
            if val == "--pipe" or val == "pipe":
                continue
            if val in args or val in kwargs.values():
                continue
            if val.startswith("--"):
                val = val[2:]
                if "=" in val:
                    k, v = val.split("=", 1)
                    kwargs[k] = v
                    continue
                args.append(val)
            else:
                if args:
                    kwargs[args.pop(-1)] = val
        for k, v in kwargs.items():
            if v == "--pipe" or v == "pipe":
                continue
            if k == "--pipe" or k == "pipe":
                continue
            if v.startswith('"') and v.endswith('"'):
                v = v[1:-1]
                kwargs[k] = v
            if v.startswith("'") and v.endswith("'"):
                v = v[1:-1]
                kwargs[k] = v
            if str(v).lower() == "false":
                kwargs[k] = False
            elif str(v).lower() == "true":
                kwargs[k] = True
        for a in args:
            kwargs[a] = True

        # all loaders need a path arg except anki and string
        if (
            "filetype" in kwargs
            and kwargs["filetype"] in ["anki", "string"]
            and "path" not in kwargs
        ):
            kwargs["path"] = None
        if is_verbose:
            whi(f"Arguments that will be used for parser: '{kwargs}'")
        parsed = wdoc.parse_file(**kwargs)
    else:
        parsed = fire.Fire(wdoc.parse_file)

    if isinstance(parsed, list):
        if all(not isinstance(d, dict) for d in parsed):
            parsed_d = [
                {"page_content": d.page_content, "metadata": d.metadata} for d in parsed
            ]
        else:
            parsed_d = parsed
        try:
            out = json.dumps(parsed_d, indent=2, ensure_ascii=False)
        except Exception:
            out = str(parsed_d)
    else:
        assert isinstance(parsed, str)
        out = parsed

    sys.stdout.write(out)
    sys.stdout.flush()


if __name__ == "__main__":
    cli_launcher()



"""
called at wdoc instance creation. It parsed the combined filetype
into an individual list of DocDict describing each a document (or in some cases
a list of documents for example a whole anki database).
This list is then processed in loaders.py, multithreading or multiprocessing
is used.
"""

import json
import random
import re
import shutil
import sys
import time
import uuid
from collections import Counter
from functools import cache as memoizer
from multiprocessing.context import TimeoutError as MultiprocessTimeoutError
from pathlib import Path

import dill
import rtoml
import uuid6
from beartype.typing import List, Optional, Tuple, Union
from joblib import Parallel, delayed
from langchain.docstore.document import Document
from tqdm import tqdm

from .env import WDOC_BEHAVIOR_EXCL_INCL_USELESS, WDOC_MAX_LOADER_TIMEOUT
from .flags import is_debug, is_verbose, is_piped
from .loaders import (
    load_one_doc_wrapped,
    load_youtube_playlist,
    loaders_temp_dir_file,
    markdownlink_regex,
    yt_link_regex,
)
from .logger import logger, red, whi
from .misc import (
    DocDict,
    ModelName,
    cache_dir,
    doc_loaders_cache,
    file_hasher,
    get_tkn_length,
    min_token,
    unlazyload_modules,
)
from .typechecker import optional_typecheck

assert WDOC_BEHAVIOR_EXCL_INCL_USELESS in [
    "warn",
    "crash",
], "Unexpected value of WDOC_BEHAVIOR_EXCL_INCL_USELESS"


# rules used to attribute input to proper filetype. For example
# any link containing youtube will be treated as a youtube link
inference_rules = {
    # format:
    # key is output filtype, value is list of regex that if match
    # will return the key
    # the order of the keys is important
    "youtube_playlist": ["youtube.*playlist"],
    "youtube": ["youtube", "invidi", r"youtu\."],
    "logseq_markdown": [".*logseq.*.md"],
    "txt": [".txt$", ".md$"],
    "online_pdf": ["^http.*pdf.*"],
    "pdf": [".*pdf$"],
    "url": ["^http"],
    "local_html": [r"^(?!http).*\.html?$"],
    "local_audio": [r".*(mp3|m4a|ogg|flac)$"],
    "epub": [".epub$"],
    "powerpoint": [".ppt$", ".pptx$", ".odp$"],
    "word": [".doc$", ".docx$", ".odt$"],
    "local_video": [".mp4", ".avi", ".mkv"],
    "json_entries": [".*.json"],
    "toml_entries": [".*.toml"],
}

recursive_types = [
    "recursive_paths",
    "json_entries",
    "toml_entries",
    "link_file",
    "youtube_playlist",
    "auto",
]

# compile the inference rules as regex
for k, v in inference_rules.items():
    for i, vv in enumerate(v):
        inference_rules[k][i] = re.compile(vv)


@optional_typecheck
def batch_load_doc(
    llm_name: ModelName,
    filetype: str,
    task: str,
    backend: str,
    n_jobs: int,
    **cli_kwargs,
) -> List[Document]:
    """load the input"""

    # just in case, make sure all modules are loaded
    unlazyload_modules()

    if "path" in cli_kwargs and isinstance(cli_kwargs["path"], str):
        cli_kwargs["path"] = cli_kwargs["path"].strip()

    # used to make sure all source tags were indeed parsed
    asked_source_tags = []
    if "source_tag" in cli_kwargs:
        asked_source_tags.append(cli_kwargs["source_tag"])

    # expand the list of document to load as long as there are recursive types
    to_load = [cli_kwargs.copy()]
    to_load[-1]["filetype"] = filetype.lower()
    new_doc_to_load = []
    while any(d["filetype"] in recursive_types for d in to_load):
        for ild, load_kwargs in enumerate(to_load):

            if "source_tag" in load_kwargs:
                if load_kwargs["source_tag"] not in asked_source_tags:
                    asked_source_tags.append(load_kwargs["source_tag"])

            to_load[ild]["filetype"] = to_load[ild]["filetype"].lower()
            if not ("path" in load_kwargs and load_kwargs["path"]):
                continue
            load_filetype = load_kwargs["filetype"]

            # auto parse filetype if infer
            if load_filetype == "auto":
                for k, v in inference_rules.items():
                    for vv in inference_rules[k]:
                        if vv.search(load_kwargs["path"]):
                            load_filetype = k
                            break
                    if load_filetype != "auto":
                        break
                if load_filetype == "auto":
                    try:
                        fp = Path(load_kwargs["path"])
                        if fp.exists():
                            try:
                                import magic

                                info = magic.from_file(fp).lower()
                            except Exception as err:
                                raise Exception(
                                    f"Failed to run python-magic as a last resort heuristic: '{err}'"
                                ) from err
                            if "pdf" in info:
                                load_filetype = "pdf"
                                break
                            elif "mpeg" in info:
                                load_filetype = "local_audio"
                                break
                            elif "epub" in info:
                                load_filetype = "epub"
                                break
                            else:
                                raise Exception(
                                    "No more python magic heuristics to try"
                                )
                    except Exception as err:
                        red(
                            f"Failed to detect 'auto' filetype for '{fp}' with regex and even python-magic. Error: '{err}'"
                        )

                assert (
                    load_filetype != "auto"
                ), f"Could not infer filetype of {load_kwargs['path']}. Use the 'filetype' argument."
                if load_filetype not in recursive_types:
                    to_load[ild]["filetype"] = load_filetype

            if load_filetype not in recursive_types:
                continue
            del load_kwargs["filetype"]

            if load_filetype == "recursive_paths":
                new_doc_to_load.extend(
                    parse_recursive_paths(cli_kwargs=cli_kwargs, **load_kwargs)
                )
                break

            elif load_filetype == "json_entries":
                new_doc_to_load.extend(
                    parse_json_entries(cli_kwargs=cli_kwargs, **load_kwargs)
                )
                break

            elif load_filetype == "toml_entries":
                new_doc_to_load.extend(
                    parse_toml_entries(cli_kwargs=cli_kwargs, **load_kwargs)
                )
                break

            elif load_filetype == "link_file":
                new_doc_to_load.extend(
                    parse_link_file(
                        cli_kwargs=cli_kwargs,
                        **load_kwargs,
                    )
                )
                break

            elif load_filetype == "youtube_playlist":
                new_doc_to_load.extend(
                    parse_youtube_playlist(cli_kwargs=cli_kwargs, **load_kwargs)
                )
                break

        if new_doc_to_load:
            for indtl, ndtl in enumerate(new_doc_to_load):
                assert (
                    ndtl
                ), f"Args for document #{indtl} from recursive_types '{load_filetype}' is empty."

            assert load_filetype in recursive_types
            to_load.remove(to_load[ild])
            to_load.extend(new_doc_to_load)
            new_doc_to_load = []
            continue

    try:
        to_load = [d if isinstance(d, DocDict) else DocDict(d) for d in to_load]
    except Exception as err:
        raise Exception(f"Expected to have only DocDict at this point: {err}'")

    for d in to_load:
        assert d, "One document to load has no arguments, double check your inputs"

    # if is not a file, check if it's not just because spaces are not escaped or something dumb like that
    for idoc, doc in enumerate(to_load):
        if "path" not in doc:
            continue

        p = Path(doc["path"])
        if p.exists():
            continue
        alternatives = [
            Path(p.expanduser()),
        ]
        for ialt, alt in enumerate(alternatives):
            if Path(alt).exists():
                if isinstance(doc["path"], Path):
                    doc["path"] = alt
                elif isinstance(doc["path"], str):
                    doc["path"] = str(alt.absolute())
                else:
                    raise ValueError(
                        f"At this point path should only be str or Path: {doc['path']}"
                    )
                break

    # remove duplicate documents
    temp = []
    for d in to_load:
        if d in temp:
            red(f"Removed document {d} (duplicate)")
        else:
            temp.append(d)
    to_load = temp

    assert to_load, f"empty list of documents to load from filetype '{filetype}'"

    # look for unexpected keys that are not relevant to doc loading, because that would
    # skip the cache
    all_unexp_keys = set()
    for doc in to_load:
        to_del = [k for k in doc if k not in DocDict.allowed_keys]
        for k in to_del:
            all_unexp_keys.add(k)
            del doc[k]
            assert k not in [
                "include",
                "exclude",
            ], "Include or exclude arguments should be reomved at this point"

    if "summar" not in task:
        # shuffle the list of files to load to make
        # the hashing progress bar more representative
        to_load = sorted(to_load, key=lambda x: random.random())

    # store the file hash in the doc kwarg
    doc_hashes = Parallel(
        n_jobs=-1 if len(to_load) > 1 else 1,
        backend=backend,
        verbose=0 if not is_verbose else 51,
    )(
        delayed(file_hasher)(doc=doc)
        for doc in tqdm(
            to_load,
            desc="Hashing files",
            unit="doc",
            colour="magenta",
            disable=len(to_load) <= 10_000 or is_piped,
        )
    )
    for i, h in enumerate(doc_hashes):
        to_load[i]["file_hash"] = doc_hashes[i]

    if "summar" not in task:
        # shuffle the list of files again to be random but deterministic:
        # keeping only the digits of each hash, then multiplying by the
        # index of the filetype by size. This makes sure the doc dicts are
        # sorted by increasing order of filetype frequency, so if there's
        # an error with the code of this filetype of its args the user knows
        # it quickly instead of after waiting a super long time
        bins = {}
        for d in to_load:
            if d["filetype"] not in bins:
                bins[d["filetype"]] = 1
            else:
                bins[d["filetype"]] += 1
        sorted_filetypes = sorted(bins.keys(), key=lambda x: bins[x])

        @optional_typecheck
        def deterministic_sorter(doc_dict: DocDict) -> int:
            h = doc_dict["file_hash"]
            h2 = "".join(filter(str.isdigit, h))
            h_ints = int(h2) if h2.isdigit() else int(random.random() * 1000)
            h_ordered = h_ints * (
                10 ** (sorted_filetypes.index(doc_dict["filetype"]) + 1)
            )
            return h_ordered

        to_load = sorted(
            to_load,
            key=deterministic_sorter,
        )

    # load_functions are slow to load so loading them here in advance for every file
    if any(("load_functions" in doc and doc["load_functions"]) for doc in to_load):
        whi("Preloading load_functions")
        for idoc, doc in enumerate(to_load):
            if "load_functions" in doc:
                if doc["load_functions"]:
                    to_load[idoc]["load_functions"] = parse_load_functions(
                        tuple(doc["load_functions"])
                    )

    to_load = list(set(to_load))  # remove duplicates docdicts

    if len(to_load) > 1:
        for tl in to_load:
            assert (
                tl["filetype"] != "text"
            ), "You shouldn't not be using filetype 'text' with other kind of documents normally. Please open an issue on github and explain me your usecase to see how I can fix that for you!"

    # dir name where to store temporary files
    load_temp_name = "file_load_" + str(uuid6.uuid6())
    # delete previous temp dir if it's several days old
    for f in cache_dir.iterdir():
        f = f.resolve()
        if (
            f.is_dir()
            and f.name.startswith("file_load_")
            and (abs(time.time() - f.stat().st_mtime) > 2 * 86400)
        ):
            assert str(cache_dir.absolute()) in str(f.absolute())
            shutil.rmtree(f)
    temp_dir = cache_dir / load_temp_name
    temp_dir.mkdir(exist_ok=False)
    loaders_temp_dir_file.write_text(str(temp_dir.absolute().resolve()))

    loader_max_timeout = WDOC_MAX_LOADER_TIMEOUT
    if loader_max_timeout <= 0:
        loader_max_timeout = None

    sharedmem = None
    if len(to_load) == 1:
        n_jobs = 1
        to_load[0][
            "loading_failure"
        ] = "crash"  # crash if loading fails when only one document is to be loaded and fails anyway
    else:
        if backend == "loky":
            if is_verbose:
                red("Using loky backend so not using 'require=sharedmem'")
        else:
            sharedmem = "sharedmem"

    # # early stopping if all the documents of a recursive filetype failed
    # expected_recur_nb = {}
    # for d in to_load:
    #     if d["recur_parent_id"] not in expected_recur_nb:
    #         expected_recur_nb[d["recur_parent_id"]] = 1
    #     else:
    #         expected_recur_nb[d["recur_parent_id"]] += 1
    # found_recur_nb = {k: 0 for k in expected_recur_nb.keys()}

    docs = []
    t_load = time.time()
    try:
        generator_doc_lists = Parallel(
            n_jobs=n_jobs,
            backend=backend,
            verbose=0 if not is_verbose else 51,
            timeout=loader_max_timeout,
            return_as="generator",  # try to reduce memory footprint
            require=sharedmem,
        )(
            delayed(load_one_doc_wrapped)(
                llm_name=llm_name,
                task=task,
                temp_dir=temp_dir,
                **d,
            )
            for d in tqdm(
                to_load,
                desc="Loading",
                unit="doc",
                colour="magenta",
                disable=is_piped,
            )
        )
        doc_lists = []
        for idoc, o in enumerate(generator_doc_lists):
            doc_lists.append(o)
            # # detect errors, used for early stopping if all the doc from the same recursive parent failed:
            # if isinstance(o, str):
            #     d = to_load[idoc]
            #     if "recur_parent_id" in d:
            #         assert d["recur_parent_id"] in expected_recur_nb, expected_recur_nb
            #         found_recur_nb[d["recur_parent_id"]] -= 1
            #         if found_recur_nb[d["recur_parent_id"]] <= 0 and expected_recur_nb[d['recur_parent_id']] > 1:
            #             mess = f"All document from a recursive file with parent id {d['recur_parent_id']} failed so crashing."
            #             mess += "\nFailed documents:"
            #             cnt = 0
            #             for ifd, fd in enumerate(to_load):
            #                 if fd['recur_parent_id'] == d["recur_parent_id"]:
            #                     cnt += 1
            #                     er = doc_lists[ifd]
            #                     assert isinstance(er, str), er
            #                     mess += f"\n- #{cnt}: {fd}: '{er}'"
            #             mess += f"\nLatest error was: '{o}'"
            #             raise Exception(red(mess))
    except MultiprocessTimeoutError as e:
        raise Exception(
            red(f"Timed out when loading batch files after {loader_max_timeout}s")
        ) from e

    # erases content that links to the loaders temporary files at startup
    loaders_temp_dir_file.write_text("")

    red(f"Done loading all {len(to_load)} documents in {time.time()-t_load:.2f}s")
    missing_docargs = []
    for idoc, d in tqdm(
        enumerate(doc_lists),
        total=len(doc_lists),
        desc="Concatenating results",
        disable=not is_verbose or is_piped,
    ):
        if isinstance(d, list):
            docs.extend(d)
        else:
            assert isinstance(d, str)
            missing_docargs.append(
                dict(to_load[idoc])
            )  # must be cast as dict to set error message
            missing_docargs[-1]["error_message"] = d
    assert not any(isinstance(d, str) for d in docs)

    if missing_docargs:
        missing_docargs = sorted(
            missing_docargs, key=lambda x: json.dumps(x, ensure_ascii=False)
        )
        red(f"Number of failed documents: {len(missing_docargs)}:")
    else:
        red("No document failed to load!")

    if len(missing_docargs) == len(doc_lists):
        raise Exception("All documents failed to load. The errors appear above.")

    if asked_source_tags:
        no_st = 0
        st = {t: 0 for t in asked_source_tags}
        extra = {}
        for doc in docs:
            if "source_tag" in doc.metadata:
                s = doc.metadata["source_tag"]
                if s not in st:
                    if s in extra:
                        extra[s] += 1
                    else:
                        extra[s] = 1
                else:
                    st[s] += 1
            else:
                no_st += 1
        should_crash = False
        red("Found the following source_tag after loading all documents:")
        for n, s in st.items():
            red(f"- {s}: {n}")
            if n == 0:
                should_crash = True
        if extra:
            red("Found the following EXTRA source_tag after loading all documents:")
            red("(This can happen after merging identical documents though)")
            should_crash = True
            for n, s in extra.items():
                red(f"- {s}: {n}")
        red(f"Found {no_st} documents with no source_tag")

        if should_crash:
            red(
                "Something might have gone wrong given those source tags.\nAnswer 'crash' to crash, 'd' to debug, anything else to continue."
            )
            ans = input(">")
            if ans == "d":
                breakpoint()
            elif ans == "crash":
                raise Exception("Probable error given the source tags")
            else:
                pass

    # smart deduplication before embedding:
    # find the document with the same content_hash, merge their metadata and keep only one
    if "summar" not in task and len(docs) > 1:
        red("Deduplicating...")
        whi("Getting all hash")
        content_hash = [d.metadata["content_hash"] for d in docs]
        whi("Counting them")
        counts = Counter(content_hash)
        dupes = set()
        [dupes.add(h) for h, c in counts.items() if c > 1]
        deduped = {}
        lenbefore = len(docs)
        for idoc, doc in enumerate(
            tqdm(docs, desc="Deduplicating", unit="doc", disable=is_piped)
        ):
            ch = doc.metadata["content_hash"]
            if not dupes:
                whi("No duplicates!")
                break
            if ch in deduped:
                assert doc.page_content == deduped[ch].page_content
                for k, v in doc.metadata.items():
                    if "hash" in k:
                        continue
                    elif k == "source_tag":
                        if "source_tag" in deduped[ch].metadata:
                            deduped[ch].metadata[k] += " " + v
                        else:
                            deduped[ch].metadata[k] += v
                        continue
                    elif k in deduped[ch].metadata:
                        if v == deduped[ch].metadata[k]:
                            continue
                    elif k not in deduped[ch].metadata:
                        deduped[ch].metadata[k] = v
                    elif isinstance(v, list) and isinstance(
                        deduped[ch].metadata[k], list
                    ):
                        deduped[ch].metadata[k] += deduped[ch].metadata[k]
                    elif is_verbose:
                        red(f"UNEXPECTED METADATA TYPE: '{k}:{v}'")
                docs[idoc] = None

            if ch in dupes:
                deduped[ch] = doc
                docs[idoc] = None
                assert counts[ch] > 1, doc
                counts[ch] -= 1
                if counts[ch] == 1:
                    dupes.remove(ch)
        if deduped:
            assert None in docs
        assert not dupes, dupes
        docs = [d for d in docs if d is not None]
        if deduped:
            docs += list(deduped.values())
        assert (
            len(docs) <= lenbefore
        ), f"Removing duplicates seems to have added documents: {lenbefore} -> {len(docs)}. Something went wrong."

    assert docs, "No documents were succesfully loaded!"

    # delete temp dir
    shutil.rmtree(temp_dir)
    assert not temp_dir.exists()

    return docs


@optional_typecheck
def parse_recursive_paths(
    cli_kwargs: dict,
    path: Union[str, Path],
    pattern: str,
    recursed_filetype: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    **extra_args,
) -> List[Union[DocDict, dict]]:
    whi(f"Parsing recursive load_filetype: '{path}'")
    assert recursed_filetype not in [
        "recursive_paths",
        "json_entries",
        "youtube",
        "anki",
    ], "'recursed_filetype' cannot be 'recursive_paths', 'json_entries', 'anki' or 'youtube'"

    if not Path(path).exists() and Path(path.replace(r"\ ", " ")).exists():
        logger.info(r"File was not found so replaced '\ ' by ' '")
        path = path.replace(r"\ ", " ")
    assert Path(path).exists(), f"not found: {path}"
    doclist = [p for p in Path(path).rglob(pattern)]
    assert doclist, f"No document found by pattern {pattern}"
    doclist = [str(p).strip() for p in doclist if p.is_file()]
    assert doclist, "No document after filtering by file"
    doclist = [p for p in doclist if p]
    assert doclist, "No document after removing nonemtpy"
    doclist = [p[1:].strip() if p.startswith("-") else p.strip() for p in doclist]
    recur_parent_id = str(uuid.uuid4())

    if include:
        for iinc, inc in enumerate(include):
            if isinstance(inc, str):
                if inc == inc.lower():
                    inc = re.compile(inc, flags=re.IGNORECASE)
                else:
                    inc = re.compile(inc)
                include[iinc] = inc
        ndoclist = len(doclist)
        doclist = [d for d in doclist if any(inc.search(d) for inc in include)]
        if not len(doclist) < ndoclist:
            mess = f"Include rules were useless and didn't filter out anything.\nInclude rules: '{include}'"
            if WDOC_BEHAVIOR_EXCL_INCL_USELESS == "warn":
                red(mess)
            elif WDOC_BEHAVIOR_EXCL_INCL_USELESS == "crash":
                raise Exception(red(mess))

    if exclude:
        for iexc, exc in enumerate(exclude):
            if isinstance(exc, str):
                if exc == exc.lower():
                    exc = re.compile(exc, flags=re.IGNORECASE)
                else:
                    exc = re.compile(exc)
                exclude[iexc] = exc
            ndoclist = len(doclist)
            doclist = [d for d in doclist if not exc.search(d)]
            if not len(doclist) < ndoclist:
                mess = f"Exclude rule '{exc}' was useless and didn't filter out anything.\nExclude rules: '{exclude}'"
                if WDOC_BEHAVIOR_EXCL_INCL_USELESS == "warn":
                    red(mess)
                elif WDOC_BEHAVIOR_EXCL_INCL_USELESS == "crash":
                    raise Exception(red(mess))

    for i, d in enumerate(doclist):
        doc_kwargs = cli_kwargs.copy()
        doc_kwargs["path"] = d
        doc_kwargs["filetype"] = recursed_filetype
        doc_kwargs.update(extra_args)
        doc_kwargs["recur_parent_id"] = recur_parent_id
        if doc_kwargs["filetype"] not in recursive_types:
            doclist[i] = DocDict(doc_kwargs)
        else:
            doclist[i] = doc_kwargs
    return doclist


@optional_typecheck
def parse_json_entries(
    cli_kwargs: dict,
    path: Union[str, Path],
    **extra_args,
) -> List[Union[DocDict, dict]]:
    whi(f"Loading json_entries: '{path}'")
    doclist = str(Path(path).read_text()).splitlines()
    doclist = [p[1:].strip() if p.startswith("-") else p.strip() for p in doclist]
    doclist = [
        p.strip() for p in doclist if p.strip() and not p.strip().startswith("#")
    ]
    recur_parent_id = str(uuid.uuid4())

    for i, d in enumerate(doclist):
        meta = cli_kwargs.copy()
        meta["filetype"] = "auto"
        meta.update(json.loads(d.strip()))
        for k, v in cli_kwargs.copy().items():
            if k not in meta:
                meta[k] = v
        if meta["path"] == path:
            del meta["path"]
        meta.update(extra_args)
        meta["recur_parent_id"] = recur_parent_id
        if meta["filetype"] not in recursive_types:
            doclist[i] = DocDict(meta)
        else:
            doclist[i] = meta
    return doclist


@optional_typecheck
def parse_toml_entries(
    cli_kwargs: dict,
    path: Union[str, Path],
    **extra_args,
) -> List[Union[DocDict, dict]]:
    whi(f"Loading toml_entries: '{path}'")
    content = rtoml.load(toml=Path(path))
    assert isinstance(content, dict)
    doclist = list(content.values())
    assert all(len(d) == 1 for d in doclist)
    doclist = [d[0] for d in doclist]
    assert all(isinstance(d, dict) for d in doclist)
    recur_parent_id = str(uuid.uuid4())

    for i, d in enumerate(doclist):
        meta = cli_kwargs.copy()
        meta["filetype"] = "auto"
        meta.update(d)
        for k, v in cli_kwargs.items():
            if k not in meta:
                meta[k] = v
        if meta["path"] == path:
            del meta["path"]
        meta.update(extra_args)
        meta["recur_parent_id"] = recur_parent_id
        if meta["filetype"] not in recursive_types:
            doclist[i] = DocDict(meta)
        else:
            doclist[i] = meta
    return doclist


@optional_typecheck
def parse_link_file(
    cli_kwargs: dict,
    path: Union[str, Path],
    **extra_args,
) -> List[DocDict]:
    whi(f"Loading link_file: '{path}'")
    doclist = str(Path(path).read_text()).splitlines()
    doclist = [p[1:].strip() if p.startswith("-") else p.strip() for p in doclist]
    doclist = [
        p.strip()
        for p in doclist
        if p.strip() and not p.strip().startswith("#") and "http" in p
    ]
    doclist = [
        matched.group(0)
        for d in doclist
        if (matched := markdownlink_regex.search(d).strip())
    ]

    recur_parent_id = str(uuid.uuid4())
    for i, d in enumerate(doclist):
        assert "http" in d, f"Link does not appear to be a link: '{d}'"
        doc_kwargs = cli_kwargs.copy()
        doc_kwargs["path"] = d
        doc_kwargs["subitem_link"] = d
        doc_kwargs["filetype"] = "auto"
        doc_kwargs.update(extra_args)
        doc_kwargs["recur_parent_id"] = recur_parent_id
        doclist[i] = DocDict(doc_kwargs)
    return doclist


@optional_typecheck
def parse_youtube_playlist(
    cli_kwargs: dict,
    path: Union[str, Path],
    **extra_args,
) -> List[DocDict]:
    if "\\" in path:
        red(f"Removed backslash found in '{path}'")
        path = path.replace("\\", "")
    whi(f"Loading youtube playlist: '{path}'")
    video = load_youtube_playlist(path)

    playlist_title = video["title"].strip().replace("\n", "")
    assert (
        "duration" not in video
    ), f'"duration" found when loading youtube playlist. This might not be a playlist: {path}'
    doclist = [ent["webpage_url"] for ent in video["entries"]]
    doclist = [li for li in doclist if yt_link_regex.search(li)]

    recur_parent_id = str(uuid.uuid4())
    for i, d in enumerate(doclist):
        assert "http" in d, f"Link does not appear to be a link: '{d}'"
        doc_kwargs = cli_kwargs.copy()
        doc_kwargs["path"] = d
        doc_kwargs["filetype"] = "youtube"
        doc_kwargs["subitem_link"] = d
        doc_kwargs.update(extra_args)
        doc_kwargs["recur_parent_id"] = recur_parent_id
        doclist[i] = DocDict(doc_kwargs)

    assert doclist, f"No video found in youtube playlist: {path}"
    return doclist


@optional_typecheck
@memoizer
def parse_load_functions(load_functions: Tuple[str, ...]) -> bytes:
    load_functions = list(load_functions)
    assert isinstance(load_functions, list), "load_functions must be a list"
    assert all(
        isinstance(lf, str) for lf in load_functions
    ), "load_functions elements must be strings"

    try:
        for ilf, lf in enumerate(load_functions):
            load_functions[ilf] = eval(lf)
    except Exception as err:
        raise Exception(f"Error when evaluating load_functions #{ilf}: {lf} '{err}'")
    load_functions = tuple(load_functions)
    pickled = dill.dumps(load_functions)
    return pickled



# Using a __init__.py to force the order of initialization:
# 1. load the env variables
# 2. Enable (or not) import tricks for faster startup time
from . import (
    batch_file_loader,
    customs,
    env,
    flags,
    import_tricks,
    loaders,
    misc,
    prompts,
    tasks,
)

__all__ = [
    "flags",
    "env",
    "batch_file_loader",
    "loaders",
    "misc",
    "prompts",
    "tasks",
    "customs",
    "import_tricks",
]



"""
Simple script to create a filtered deck from a wdoc search
"""

import os
from typing import List, Tuple

import fire
from langchain.docstore.document import Document
from loguru import logger
from py_ankiconnect import PyAnkiconnect

from wdoc import wdoc
from wdoc.utils.env import WDOC_DEFAULT_MODELNAME
from wdoc.utils.typechecker import optional_typecheck

VERSION = "1.1"

# logger
logger.add(
    "logs.txt",
    rotation="100MB",
    retention=5,
    format="{time} {level} {thread} FilteredDeckCreator {process} {function} {line} {message}",
    level="DEBUG",
    enqueue=True,
    colorize=False,
)


def p(*args):
    "simple way to log to logs.txt and to print"
    print(*args)
    logger.info(*args)


# default anki host
if "PYTHON_PY_ANKICONNECT_HOST" in os.environ:
    host = os.environ["PYTHON_PY_ANKICONNECT_HOST"]
else:
    host: str = "http://localhost"
if "PYTHON_PY_ANKICONNECT_PORT" in os.environ:
    port = int(os.environ["PYTHON_PY_ANKICONNECT_PORT"])
else:
    port = 8775

akc = PyAnkiconnect(
    default_host=host,
    default_port=port,
)


@optional_typecheck
class FilteredDeckCreator:
    def __init__(
        self,
        deckname: str,  # will have 'wdoc_filtered_deck::' prepended to it
        query: str,
        task: str = "search",  # could also be 'query', would be more expensive but would use the large LLM
        filtered_deck_query: str = "",
        new_tag: str = "",  # will have 'wdoc_filtered_deck::' prepended to each space separated tag
        reschedule: bool = False,
        sort_order: int = 8,
        create_empty: bool = False,
        query_eval_modelname: str = WDOC_DEFAULT_MODELNAME,  # by default, use the same model as we use normally for querying
        **kwargs,
    ) -> None:
        akc("sync")
        p(f"Started with query '{query}'")
        deckname = "wdoc_filtered_deck::" + deckname
        decknames = akc("deckNames")
        assert (
            deckname not in decknames
        ), f"Deckname {deckname} is already used. You have to delete it manually"

        instance = wdoc(
            query_eval_modelname=query_eval_modelname,
            task=task,
            import_mode=True,
            query=query,
            **kwargs,
        )
        found = instance.query_task(query=query)

        if "relevant_filtered_docs" in found:
            docs = found["relevant_filtered_docs"]
        else:
            docs = found["filtered_docs"]
        nids, cids = self.find_anki_docs(
            docs=docs,
        )

        query = filtered_deck_query
        query += " (nid:" + ",".join([str(n) for n in nids]) + ")"

        if new_tag:
            if " " not in new_tag:
                new_tag = "wdoc_filtered_deck::" + new_tag
            else:
                new_tag = new_tag.replace(" ", " wdoc_filtered_deck::")
            p(f"Adding tags {new_tag} to nids:{nids}")
            akc("addTags", notes=nids, tags=new_tag)
            p("Done adding tags")

        p("Creating filtered deck")
        self.create_filtered_deck(
            deckname=deckname,
            search_query=query,
            gather_count=len(cids) + 1,
            reschedule=reschedule,
            sort_order=sort_order,
            create_empty=create_empty,
        )
        p(f"Done, created filtered deck {deckname}")
        akc("sync")

    @classmethod
    def find_anki_docs(self, docs: List[Document]) -> Tuple[List[int], List[int]]:
        """
        goes through the metadata of each langchain Document to find which
        correspond to anki cards.
        Note that the order will NOT be respected (i.e. nids, cids and documents
        will all have their own order and length)
        Returns [ nids, cids ]
        """
        anki_docs = []
        for doc in docs:
            if any("anki_" in k for k in doc.metadata.keys()):
                anki_docs.append(doc)
                continue
        assert anki_docs, "No anki notes found in documents"

        present_anki_docs = []
        cids = []
        nids = []
        for idoc, doc in enumerate(anki_docs):
            if "anki_cid" in doc.metadata:
                cid = int(doc.metadata["anki_cid"])
                if akc("findCards", query=f"cid:{cid}"):
                    present_anki_docs.append(doc)
                    cids.append(int(cid))
            if "anki_nid" in doc.metadata:
                nid = int(doc.metadata["anki_nid"])
                nids.append(nid)
                temp_cids = [int(c) for c in akc("findCards", query=f"nid:{nid}")]
                if temp_cids:
                    present_anki_docs.append(doc)
                    cids.extend(temp_cids)

        assert present_anki_docs, "No anki notes after filtering"
        assert cids, "No cids found"
        cids = list(set(cids))
        nids = list(set(nids))
        p(f"Found cids:{cids}")

        p(f"Ratio of cid per document: {len(cids)/len(docs):.4f}")
        p(f"Ratio of nid per document: {len(nids)/len(docs):.4f}")

        return nids, cids

    @classmethod
    def create_filtered_deck(
        self,
        deckname: str,
        search_query: str,
        gather_count: int = 100,
        reschedule: bool = False,
        sort_order: int = 8,
        create_empty: bool = False,
    ) -> None:
        logger.info(
            f"Creating filtered deck {deckname}, {search_query}, {gather_count}, {reschedule}, {sort_order}, {create_empty}"
        )
        akc(
            action="createFilteredDeck",
            newDeckName=deckname,
            searchQuery=search_query,
            gatherCount=gather_count,
            reschedule=reschedule,
            sortOrder=sort_order,
            createEmpty=create_empty,
        )


if __name__ == "__main__":
    fire.Fire(FilteredDeckCreator)
